from akasha.utils.base import (
    DEFAULT_MODEL,
    DEFAULT_MAX_OUTPUT_TOKENS,
    DEFAULT_MAX_INPUT_TOKENS,
)
from .ask import ask, _retri_max_texts
from akasha.utils.prompts.gen_prompt import (
    default_ask_prompt,
    default_conclusion_prompt,
    format_sys_prompt,
)
import time
import datetime
from akasha.utils.prompts.format import handle_params, handle_metrics, handle_table
from akasha.helper.base import get_doc_length
from akasha.helper.run_llm import call_model, call_batch_model, check_relevant_answer
from akasha.helper.web_engine import load_docs_from_webengine

from akasha.helper.token_counter import myTokenizer


class websearch(ask):
    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        max_input_tokens: int = DEFAULT_MAX_INPUT_TOKENS,
        max_output_tokens: int = DEFAULT_MAX_OUTPUT_TOKENS,
        temperature: float = 0.0,
        prompt_format_type: str = "auto",
        search_engine: str = "wiki",
        search_num: int = 5,
        language: str = "ch",
        record_exp: str = "",
        system_prompt: str = "",
        keep_logs: bool = False,
        verbose: bool = False,
        stream: bool = False,
        env_file: str = "",
    ):
        """websearch class will search the user prompt in the web and based on the results to answer the question.

        Args:
            model (str, optional): _description_. Defaults to DEFAULT_MODEL.
            max_input_tokens (int, optional): _description_. Defaults to DEFAULT_MAX_INPUT_TOKENS.
            max_output_tokens (int, optional): _description_. Defaults to DEFAULT_MAX_OUTPUT_TOKENS.
            temperature (float, optional): _description_. Defaults to 0.0.
            language (str, optional): _description_. Defaults to "ch".
            search_engine (str, optional): the search api methods, includes "serper", "brave","wiki" . Defaults to "wiki".
            search_num (int, optional): the number of search results. Defaults to 5.
            record_exp (str, optional): _description_. Defaults to "".
            system_prompt (str, optional): _description_. Defaults to "".
            keep_logs (bool, optional): _description_. Defaults to False.
            verbose (bool, optional): _description_. Defaults to False.
            env_file (str, optional): _description_. Defaults to "".
        """
        super().__init__(
            model=model,
            max_input_tokens=max_input_tokens,
            max_output_tokens=max_output_tokens,
            temperature=temperature,
            prompt_format_type=prompt_format_type,
            language=language,
            record_exp=record_exp,
            system_prompt=system_prompt,
            keep_logs=keep_logs,
            verbose=verbose,
            stream=stream,
            env_file=env_file,
        )
        self.prompt = ""
        self.response = ""
        self.search_engine = search_engine.lower()
        self.search_num = search_num
        self.docs = []
        self.prompt_tokens, self.prompt_length = 0, 0
        self.doc_tokens, self.doc_length = 0, 0

        ## set default RAG prompt ##
        if self.system_prompt.replace(" ", "") == "":
            self.system_prompt = default_ask_prompt(self.language)

    def _display_info(self, batch: int = 1) -> bool:
        if self.verbose is False:
            return False

        print(f"Model: {self.model}, Temperature: {self.temperature}")
        print(f"Search engine: {self.search_engine}, Search num: {self.search_num}")
        print(
            f"Prompt format type: {self.prompt_format_type}, Max input tokens: {self.max_input_tokens}"
        )
        print(
            f"Prompt tokens: {self.prompt_tokens}, Prompt length: {self.prompt_length}"
        )
        print(f"Doc tokens: {self.doc_tokens}, Doc length: {self.doc_length}\n\n")

        return True

    def _add_basic_log(
        self,
        timestamp: str,
        fn_type: str,
    ) -> bool:
        if super()._add_basic_log(timestamp, fn_type) is False:
            return False

        self.logs[timestamp]["prompt"] = self.prompt
        self.logs[timestamp]["search_engine"] = self.search_engine
        self.logs[timestamp]["search_num"] = self.search_num
        self.logs[timestamp].pop("history_messages", None)
        return True

    def _add_result_log(self, timestamp, time) -> bool:
        if super()._add_result_log(timestamp, time) is False:
            return False

        ### add token information ###
        self.logs[timestamp]["response"] = self.response
        self.logs[timestamp]["prompt_tokens"] = self.prompt_tokens
        self.logs[timestamp]["prompt_length"] = self.prompt_length
        self.logs[timestamp]["doc_tokens"] = self.doc_tokens
        self.logs[timestamp]["doc_length"] = self.doc_length
        return True

    def _upload_logs(self, tot_time: float, doc_len: int, doc_tokens: int) -> str:
        """_summary_

        Args:
            tot_time (float): _description_
            doc_len (int): _description_
            doc_tokens (int): _description_
        """
        if self.record_exp == "":
            return "no record_exp assigned, so no logs uploaded"

        params = handle_params(
            self.model,
            self.language,
            "",
        )
        params["search_engine"] = self.search_engine
        params["search_num"] = self.search_num
        metrics = handle_metrics(doc_len, tot_time, doc_tokens)
        table = handle_table(self.prompt, self.docs, self.response)
        from akasha.utils.upload import aiido_upload

        aiido_upload(self.record_exp, params, metrics, table)

        return "logs uploaded"

    def __call__(self, prompt: str, **kwargs):
        """search the prompt in the web and based on the results to answer the question.

        Args:
            prompt (str): the user input prompt

        Returns:
            str: the response of the prompt
        """
        self._set_model(**kwargs)
        self._change_variables(**kwargs)
        self.prompt = prompt
        self.response = ""

        start_time = time.time()
        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        self._add_basic_log(timestamp, "websearch")
        self.docs = load_docs_from_webengine(
            self.prompt,
            self.search_engine,
            self.search_num,
            self.language,
            self.env_file,
        )

        ### check if prompt <= max_input_tokens ###
        tot_prompts = self.prompt + self.system_prompt
        self.prompt_length = get_doc_length(self.language, tot_prompts)
        self.prompt_tokens = myTokenizer.compute_tokens(tot_prompts, self.model)

        if self.prompt_tokens > self.max_input_tokens:
            print("\n\nThe tokens of prompt is larger than max_input_tokens.\n\n")
            raise ValueError("The tokens of prompt is larger than max_input_tokens.")

        ### separate documents and count tokens ###
        cur_documents, self.doc_tokens = self._separate_docs()
        self.doc_length = get_doc_length(self.language, "".join(cur_documents))

        prod_sys_prompts = self._process_batch_prompts(cur_documents)

        self._display_info(
            len(cur_documents)
        )  # display the information of the parameters
        self._display_docs()

        ### start to ask llm ###
        if len(cur_documents) > 1:
            ## call batch model ##
            batch_responses = call_batch_model(
                self.model_obj,
                prod_sys_prompts,
            )
            fnl_conclusion_prompt = default_conclusion_prompt(prompt, self.language)
            ## check relevant answer if batch_responses > 10 ##
            if len(batch_responses) > 10:
                batch_responses = check_relevant_answer(
                    self.model_obj,
                    batch_responses,
                    self.prompt,
                    self.prompt_format_type,
                )

            batch_responses, cur_len = _retri_max_texts(
                batch_responses,
                self.max_input_tokens
                - myTokenizer.compute_tokens(fnl_conclusion_prompt, self.model),
                self.model,
            )

            fnl_input = format_sys_prompt(
                fnl_conclusion_prompt,
                "\n\n".join(batch_responses),
                self.prompt_format_type,
                self.model,
            )

            if self.stream:
                return self._display_stream(
                    fnl_input,
                )

            self.response = call_model(self.model_obj, fnl_input)

        else:
            if self.stream:
                return self._display_stream(prod_sys_prompts[0])

            self.response = call_model(self.model_obj, prod_sys_prompts[0])

        end_time = time.time()
        self._add_result_log(timestamp, end_time - start_time)

        self._upload_logs(end_time - start_time, self.doc_length, self.doc_tokens)
        return self.response



import streamlit as st
import sys
import akasha
from pathlib import Path
import datetime

sys.path.append("../")


def summary_page():
    if "prompt_list" not in st.session_state:
        st.session_state.prompt_list = []

    st.title("Summarize")
    st.markdown("##")
    st.markdown("##")

    response_board, para_set = st.columns([2, 1])
    with para_set:
        sum_t, sum_l = st.columns([1, 1])
        with sum_t:
            summary_type = st.selectbox(
                "Summary Type",
                ["map_reduce", "refine"],
                index=0,
                help="map_reduce is faster but may lack clarity and detail, refine is slower but offers a comprehensive understanding of the content.",
            )
        with sum_l:
            summary_len = st.number_input(
                "Summary Length",
                value=300,
                min_value=100,
                max_value=1000,
                step=10,
                help="The length of the output summary you want LLM to generate.",
            )
        uploaded_file = st.file_uploader("Upload a file", type=["txt", "pdf", "docx"])

        sb1, sb2 = st.columns([1, 1])
        with sb1:
            if st.button(
                "Clear",
                type="primary",
                use_container_width=True,
                help="Clear the prompt and response.",
            ):
                st.session_state.prompt_list = []
                st.session_state.response_list = []

                st.rerun()
        with sb2:
            if st.button("Submit", type="primary", use_container_width=True):
                if uploaded_file is not None:
                    # Get the file path
                    file_date = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
                    file_name = uploaded_file.name
                    bytes_data = uploaded_file.read()
                    with open(file_date + file_name, "wb") as f:
                        f.write(bytes_data)

                    if not isinstance(st.session_state.akasha_obj, akasha.summary):
                        st.session_state.akasha_obj = akasha.summary(
                            chunk_size=st.session_state.chunksize,
                            chunk_overlap=40,
                            model=st.session_state.model,
                            language="ch",
                            verbose=True,
                            record_exp="",
                            max_input_tokens=st.session_state.max_input_tokens,
                            temperature=st.session_state.temperature,
                            env_file=st.session_state.env_path,
                        )

                    ans = st.session_state.akasha_obj(
                        file_date + file_name,
                        sum_type=summary_type,
                        sum_len=summary_len,
                        chunk_size=st.session_state.chunksize,
                        max_input_tokens=st.session_state.max_input_tokens,
                        temperature=st.session_state.temperature,
                        keep_logs=True,
                        env_file=st.session_state.env_path,
                    )

                    st.session_state.prompt_list.append("summary of " + file_name)
                    st.session_state.response_list.append(ans)
                    timesp = st.session_state.akasha_obj.timestamp_list[-1]
                    if timesp in st.session_state.akasha_obj.logs:
                        st.session_state.logs[timesp] = (
                            st.session_state.akasha_obj.logs[timesp]
                        )

                    Path(file_date + file_name).unlink()
                else:
                    st.error("Please upload a file.")

    with response_board:
        for i in range(len(st.session_state.response_list)):
            with st.chat_message("user"):
                st.markdown(st.session_state.prompt_list[i])
            with st.chat_message("assistant"):
                st.markdown(st.session_state.response_list[i])
    return



from akasha.helper.handle_objects import handle_client
from pathlib import Path
from typing import Union

IMAGE_DEFAULT_MODEL = "openai:gpt-image-1"


def gen_image(
    prompt: str,
    save_path: str = "./image.png",
    model: str = IMAGE_DEFAULT_MODEL,
    size: str = "auto",
    quality: str = "auto",
    verbose: bool = False,
    env_file: str = "",
):
    """the image generate image call Model to generate image based on the user prompt.\n

    Args:
        prompt (str): user prompt
        save_path (str, optional): image save path, can be .png, .jpeg, .webp. Defaults to "./image.png".
        model (str, optional): model name. Defaults to "openai:gpt-image-1".
        size (str, optional): image size(256x256, 1024x1024...). Defaults to "auto".
        quality (str, optional): image quality. Defaults to "auto".
        verbose (bool, optional): whether to print the response. Defaults to False.
        env_file (str, optional): env file path. Defaults to "".
    """

    model_obj = handle_client(model, env_file)

    print(
        f"Generating, may take some time...\n\nmodel: {model}, save_path: {save_path}"
    )
    ret = model_obj.generate(
        prompt=prompt,
        save_path=save_path,
        size=size,
        quality=quality,
        moderation="auto",
        background="auto",
    )

    return ret


def edit_image(
    prompt: str,
    images: Union[list[str], str, Path],
    save_path: str = "./image.png",
    model: str = IMAGE_DEFAULT_MODEL,
    size: str = "auto",
    quality: str = "auto",
    verbose: bool = False,
    env_file: str = "",
):
    """the image generate image call Model to generate image based on the user prompt.\n

    Args:
        prompt (str): user prompt
        images (list[str]): the list of image path(string) you want to use as source.
        save_path (str, optional): image save path, can be .png, .jpeg, .webp. Defaults to "./image.png".
        model (str, optional): model name. Defaults to "openai:gpt-image-1".
        size (str, optional): image size(256x256, 1024x1024...). Defaults to "auto".
        quality (str, optional): image quality. Defaults to "auto".
        verbose (bool, optional): whether to print the response. Defaults to False.
        env_file (str, optional): env file path. Defaults to "".
    """

    model_obj = handle_client(model, env_file)

    ### check if images is a list
    if not isinstance(images, list):
        images = [images]

    print(
        f"Generating, may take some time...\n\nmodel: {model}, save_path: {save_path}"
    )

    ret = model_obj.edit(
        prompt=prompt,
        images=images,
        save_path=save_path,
        size=size,
        quality=quality,
        moderation="auto",
        background="auto",
    )

    return ret



from typing import List, Any, Optional
from langchain.llms.base import LLM
from transformers import AutoTokenizer, TextStreamer

import torch
import sys
from pydantic import Field


class gptq(LLM):
    """define initials and _call function for gptq model

    Args:
        LLM (_type_): _description_

    Returns:
        _type_: _description_
    """

    max_token: int = 4096
    temperature: float = 0.01
    top_p: float = 0.95
    tokenizer: Any = Field(default=None)
    model: Any = Field(default=None)

    def __init__(
        self,
        model_name_or_path: str,
        temperature: float = 0.01,
        bit4: bool = True,
        max_token: int = 4096,
    ):
        super().__init__()
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path, use_fast=False, max_length=max_token, truncation=True
        )
        self.tokenizer.pad_token = self.tokenizer.eos_token
        self.max_token = max_token
        self.temperature = temperature
        if self.temperature == 0.0:
            self.temperature = 0.01
        if bit4 is False:
            from transformers import AutoModelForCausalLM

            self.model = AutoModelForCausalLM.from_pretrained(
                model_name_or_path,
                device_map="auto",
                torch_dtype=torch.float16,
                load_in_8bit=True,
            )
            self.model.eval()
        else:
            from auto_gptq import AutoGPTQForCausalLM

            self.model = AutoGPTQForCausalLM.from_quantized(
                model_name_or_path,
                low_cpu_mem_usage=True,
                device="cuda:0",
                use_triton=False,
                inject_fused_attention=False,
                inject_fused_mlp=False,
            )

        if torch.__version__ >= "2" and sys.platform != "win32":
            self.model = torch.compile(self.model)

    @property
    def _llm_type(self) -> str:
        return "gptq: model"

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        input_ids = self.tokenizer(
            prompt, return_tensors="pt", add_special_tokens=False
        ).input_ids.to("cuda")
        generate_input = {
            "input_ids": input_ids,
            "max_new_tokens": 1024,
            "do_sample": True,
            "top_k": 50,
            "top_p": self.top_p,
            "temperature": self.temperature,
            "repetition_penalty": 1.2,
            "eos_token_id": self.tokenizer.eos_token_id,
            "bos_token_id": self.tokenizer.bos_token_id,
            "pad_token_id": self.tokenizer.pad_token_id,
        }
        generate_ids = self.model.generate(**generate_input)
        text = self.tokenizer.decode(generate_ids[0])
        return text


class peft_Llama2(LLM):
    """define initials and _call function for llama2 peft model

    Args:
        LLM (_type_): _description_

    Returns:
        _type_: _description_
    """

    max_token: int = 2048
    temperature: float = 0.01
    top_p: float = 0.95
    tokenizer: Any = Field(default=None)
    model: Any = Field(default=None)

    def __init__(
        self, model_name_or_path: str, max_token: int = 2048, temperature: float = 0.01
    ):
        super().__init__()
        from peft import AutoPeftModelForCausalLM

        self.temperature = temperature
        if self.temperature == 0.0:
            self.temperature = 0.01
        self.max_token = max_token
        device_map = {"": 0}
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path, trust_remote_code=True
        )
        self.model = AutoPeftModelForCausalLM.from_pretrained(
            model_name_or_path + "/adapter_model",
            temperature=0.1,
            device_map=device_map,
            torch_dtype=torch.bfloat16,
        )

    @property
    def _llm_type(self) -> str:
        return "peft_Llama2:"

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        device = "cuda" if torch.cuda.is_available() else "cpu"

        inputs = self.tokenizer(prompt, return_tensors="pt").to(device)
        outputs = self.model.generate(
            input_ids=inputs["input_ids"].to(device),
            attention_mask=inputs["attention_mask"],
            max_new_tokens=512,
            pad_token_id=self.tokenizer.eos_token_id,
            temperature=self.temperature,
            do_sample=True,
        )

        result_message = self.tokenizer.decode(outputs[0], skip_special_tokens=True)

        return result_message


class TaiwanLLaMaGPTQ(LLM):
    max_token: int = 300
    temperature: float = 0.01
    top_p: float = 0.95
    tokenizer: Any = Field(default=None)
    model: Any = Field(default=None)
    streamer: Any = Field(default=None)

    def __init__(self, model_name_or_path: str, temperature: float = 0.01):
        super().__init__()
        self.temperature = temperature
        if self.temperature == 0.0:
            self.temperature = 0.01
        from auto_gptq import AutoGPTQForCausalLM

        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path,
            use_fast=True,
            max_length=4096,
            truncation=True,
            add_eos_token=True,
        )
        self.model = AutoGPTQForCausalLM.from_quantized(
            model_name_or_path,
            trust_remote_code=True,
            use_safetensors=True,
            device_map="auto",
            use_triton=False,
            strict=False,
        )

        self.streamer = TextStreamer(
            self.tokenizer, skip_prompt=True, skip_special_tokens=True
        )

    @property
    def _llm_type(self) -> str:
        return "Taiwan_LLaMa:"

    def _call(self, message: str, stop: Optional[List[str]] = None):
        prompt = message
        tokens = self.tokenizer(prompt, return_tensors="pt").input_ids
        generate_ids = self.model.generate(
            input_ids=tokens.cuda(),
            max_new_tokens=self.max_token,
            streamer=self.streamer,
            top_p=0.95,
            top_k=50,
            temperature=self.temperature,
            do_sample=True,
            eos_token_id=self.tokenizer.eos_token_id,
            bos_token_id=self.tokenizer.bos_token_id,
            pad_token_id=self.tokenizer.pad_token_id,
        )
        output = self.tokenizer.decode(generate_ids[0, len(tokens[0]) : -1]).strip()

        return output



from .retri_bm25 import myBM25Retriever
from .retri_custom import customRetriever
from .retri_knn import myKNNRetriever
from .retri_mmr import myMMRRetriever
from .retri_rerank import myRerankRetriever
from .retri_svm import mySVMRetriever
from .retri_tfidf import myTFIDFRetriever
from .retri_faiss import myFAISSRetriever

from typing import List, Union, Callable
from akasha.utils.db.db_structure import dbs
from akasha.helper.handle_objects import handle_embeddings_and_name
from langchain_core.embeddings import Embeddings
from langchain.schema import BaseRetriever


def get_retrivers(
    db: Union[dbs, list],
    embeddings: Union[Embeddings, str],
    threshold: float = 0.0,
    search_type: Union[str, Callable] = "auto",
    env_file: str = "",
) -> List[BaseRetriever]:
    """get the retrivers based on given search_type, default is auto, which contain, 'svm', 'bm25'.
       'merge' method contain 'mmr','svm','tfidf' and merge them together.

    Args:
        **db (Chromadb)**: chroma db\n
        **embeddings (Union[Embeddings, str])**: embeddings used to store vector and search documents\n
        **query (str)**: the query str used to search similar documents\n
        **threshold (float)**: (deprecated) the similarity score threshold to select documents\n
        **search_type (str)**: search type to find similar documents from db, .
            includes 'auto', 'merge', 'mmr', 'svm', 'tfidf', 'bm25'.\n
        **use_rerank (bool)**: use rerank model to search docs. Defaults to False.\n

    Returns:
        List[BaseRetriever]: selected list of retrievers that the search_type needed .
    """

    topK = 10000

    retriver_list = []

    embeddings, embed_name = handle_embeddings_and_name(embeddings, False, env_file)

    if callable(search_type):
        custom_retriver = customRetriever.from_db(
            db, embeddings, search_type, topK, threshold
        )
        retriver_list.append(custom_retriver)

    else:
        search_type = search_type.lower()

        if search_type in ["merge", "tfidf", "auto", "auto_rerank", "bm25", "rerank"]:
            docs_list = db.get_Documents()

        if search_type in ["mmr", "merge"]:
            mmr_retriver = myMMRRetriever.from_db(db, embeddings, topK, threshold)
            retriver_list.append(mmr_retriver)

        if search_type in ["svm", "merge"]:
            svm_retriver = mySVMRetriever.from_db(db, embeddings, topK, threshold)
            retriver_list.append(svm_retriver)

        if search_type in ["tfidf", "merge"]:
            tfidf_retriver = myTFIDFRetriever.from_documents(docs_list, k=topK)
            retriver_list.append(tfidf_retriver)

        if search_type in ["faiss", "FAISS", "meta", "facebook"]:
            faiss_retriver = myFAISSRetriever.from_db(db, embeddings, topK, threshold)
            retriver_list.append(faiss_retriver)

        if search_type in ["knn", "auto", "auto_rerank"]:
            knn_retriver = myKNNRetriever.from_db(db, embeddings, topK, threshold)
            retriver_list.append(knn_retriver)

        if search_type in ["bm25", "auto", "auto_rerank"]:
            bm25_retriver = myBM25Retriever.from_documents(docs_list, topK, threshold)
            retriver_list.append(bm25_retriver)

        if "rerank" in search_type:
            if ":" in search_type:
                search_type, rerank_type = search_type.split(":")
            else:
                rerank_type = "BAAI/bge-reranker-base"

            rerank_retriver = myRerankRetriever.from_documents(
                docs_list, k=topK, relevancy_threshold=threshold, model_name=rerank_type
            )
            retriver_list.append(rerank_retriver)

    if len(retriver_list) == 0:
        raise ValueError(f"cannot find search type {search_type}, end process\n")

    return retriver_list



from langchain.docstore.document import Document
from typing import Union, List


def handle_params(
    model: str,
    language: str,
    search_type: str,
    threshold: float = -1.0,
    embeddings: str = "",
    chunk_size: int = -1,
) -> dict:
    """save running parameters into dictionary in order to parse to aiido

    Args:
        **model (str)**: model name\n
        **embeddings (str)**: embedding name\n
        **chunk_size (int)**: chunk size of texts from documents\n
        **search_type (str)**: search type of finding relevant documents\n
        **threshold (float)**: only return documents that has similarity score larger than threshold\n
        **language (str)**: 'ch' for chinese and 'en' for other.\n

    Returns:
        dict: parameter dictionary
    """
    params = {}
    if model != "":
        params["model"] = model
    if embeddings != "":
        params["embeddings"] = embeddings
    if search_type != "":
        params["search_type"] = search_type
    if threshold != -1.0:
        params["threshold"] = threshold
    if language != "":
        params["language"] = language_dict[language]

    if chunk_size != -1:
        params["chunk_size"] = chunk_size
    return params


def handle_metrics(doc_length: int, time: float, tokens: int) -> dict:
    """save running metrics into dictionary in order to parse to aiido

    Args:
        **doc_length (int)**: length of texts from relevant documents  \n
        **time (float)**: total spent time\n
        **tokens (int)**: total tokens of texts from relevant documents\n

    Returns:
        dict: metric dictionary
    """
    metrics = {}

    metrics["doc_length"] = doc_length
    metrics["time"] = time
    metrics["tokens"] = tokens
    return metrics


def handle_table(prompt: str, docs: List[Union[Document, str]], response: str) -> dict:
    """save running results into dictionary in order to parse to aiido

    Args:
        **prompt (str)**: input query/question\n
        **docs (list)**: Document list and metadata\n
        **response (str)**: results from llm response\n

    Returns:
        dict: table dictionary
    """
    table = {}
    table["prompt"] = prompt
    table["response"] = response
    if len(docs) != 0:
        try:
            inputs = [doc.page_content for doc in docs]

            if "source" in docs[0].metadata and "page" in docs[0].metadata:
                # if the metadata has source and page, we can use it to show the source of the document
                metadata = [
                    doc.metadata["source"] + "    page: " + str(doc.metadata["page"])
                    for doc in docs
                ]
            elif "title" in docs[0].metadata and "url" in docs[0].metadata:
                # if the metadata has title and url, we can use it to show the source of the document
                metadata = [
                    doc.metadata["title"] + "    url: " + doc.metadata["url"]
                    for doc in docs
                ]
            else:
                raise Exception(
                    "metadata does not have source or page, or title and url"
                )
        except Exception:
            metadata = ["none" for _ in docs]
            inputs = [doc for doc in docs]
        table["inputs"] = inputs
        table["metadata"] = metadata

    else:
        table["inputs"] = ["none"]
        table["metadata"] = ["none"]

    return table


def handle_score_table(
    table: dict, bert: float, rouge: float, llm_score: float
) -> dict:
    """add each response's bert and rouge score into table dictionary

    Args:
        **table (dict)**: table dictionary that store texts data for a run of experiment.\n
        **bert (float)**: bert score\n
        **rouge (float)**: rouge score\n

    Returns:
        dict: table dictionary
    """

    table["bert"] = bert
    table["rouge"] = rouge
    table["llm_score"] = llm_score

    return table


def handle_language(language: str):
    if language not in language_dict:
        print("language not supported, use chinese as default\n\n")
        return "ch"
    else:
        return language


language_dict = {
    "en": "english",
    "ch": "traditional chinese",
    "jp": "japanese",
    "ja": "japanese",
    "zh": "traditional chinese",
    "cn": "simplified chinese",
    "de": "german",
    "el": "greek",
    "es": "spanish",
    "fr": "french",
    "it": "italian",
    "ko": "korean",
    "nl": "dutch",
    "pl": "polish",
    "pt": "portuguese",
    "ru": "russian",
    "tr": "turkish",
    "vi": "vietnamese",
    "hi": "hindi",
    "ar": "arabic",
    "th": "thai",
    "id": "indonesian",
    "no": "norwegian",
    "sv": "swedish",
    "fi": "finnish",
    "da": "danish",
    "cs": "czech",
    "hu": "hungarian",
}

websearch_language_dict = {
    "en": {"brave": "en", "serper": "en", "wiki": "en"},
    "ch": {"brave": "zh-hant", "serper": "zh-TW", "wiki": "zh-TW"},
    "jp": {"brave": "ja", "serper": "ja", "wiki": "jp"},
    "ja": {"brave": "ja", "serper": "ja", "wiki": "jp"},
}

websearch_country_dict = {
    "en": {"brave": "us", "serper": "us", "wiki": "us"},
    "ch": {"brave": "tw", "serper": "tw", "wiki": "tw"},
    "jp": {"brave": "jp", "serper": "jp", "wiki": "jp"},
    "ja": {"brave": "jp", "serper": "jp", "wiki": "jp"},
}



from google import genai
from google.genai import types

from langchain.llms.base import LLM
from typing import Dict, List, Any, Optional, Generator, Union
from pydantic import BaseModel
import os
from langchain.schema.embeddings import Embeddings
from pathlib import Path

# from pydantic import Field
import concurrent.futures


class gemini_model(LLM):
    max_token: int = 4096
    max_output_tokens: int = 1024
    temperature: float = 0.01
    top_p: float = 0.95
    history: list = []
    model_name: str = "gemini-1.5-flash"
    client: genai.Client = None
    generation_config: types.GenerateContentConfig = None

    def __init__(
        self, model_name: str, api_key: str, temperature: float = 0.0, **kwargs
    ):
        """define custom model, input func and temperature

        Args:
            **func (Callable)**: the function return response from llm\n
        """
        super().__init__()
        self.client = genai.Client(api_key=api_key)
        self.temperature = temperature
        if "max_output_tokens" in kwargs:
            self.max_output_tokens = kwargs["max_output_tokens"]

        self.generation_config = types.GenerateContentConfig(
            max_output_tokens=self.max_output_tokens,
            temperature=temperature,
        )
        self.model_name = model_name

    @property
    def _llm_type(self) -> str:
        """return llm type

        Returns:
            str: llm type
        """
        return f"gemini:{self.model_name}"

    def stream(
        self,
        prompt: Union[str, List[Dict[str, Any]]],
        stop: Optional[List[str]] = None,
        system_prompt: Union[str, None] = None,
    ) -> Generator:
        """run llm and get the stream generator

        Args:
            prompt (str): _description_

        Yields:
            Generator: _description_
        """
        if isinstance(prompt, list):
            prompt, system_prompt = check_format_prompt(prompt)

        generation_config = types.GenerateContentConfig(
            max_output_tokens=self.max_output_tokens,
            temperature=self.temperature,
            top_p=self.top_p,
            stop_sequences=stop,
            system_instruction=system_prompt,
        )
        for chunk in self.client.models.generate_content_stream(
            model=self.model_name, contents=prompt, config=generation_config
        ):
            if chunk.text:
                yield chunk.text

        return

    def _call(
        self,
        prompt: Union[str, List[Dict[str, Any]]],
        stop: Optional[List[str]] = None,
        verbose=True,
        system_prompt: Union[str, None] = None,
        response_format: Union[dict, BaseModel, None] = None,
    ) -> str:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        if isinstance(prompt, list):
            prompt, system_prompt = check_format_prompt(prompt)
        config_param = {
            "max_output_tokens": self.max_output_tokens,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "stop_sequences": stop,
            "system_instruction": system_prompt,
        }
        if response_format is not None:
            config_param["response_schema"] = list[response_format]
            config_param["response_mime_type"] = "application/json"
        generation_config = types.GenerateContentConfig(**config_param)

        ret = ""

        for chunk in self.client.models.generate_content_stream(
            model=self.model_name, contents=prompt, config=generation_config
        ):
            if chunk.text:
                print(chunk.text, end="", flush=True)
                ret += chunk.text

        return ret

    def _invoke_helper(self, args):
        messages, stop, verbose = args
        return self._call(messages, stop, verbose)

    def batch(self, prompt: List[str], stop: Optional[List[str]] = None) -> List[str]:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        # Number of threads should not exceed the number of prompts

        num_threads = min(
            len(prompt), concurrent.futures.thread.ThreadPoolExecutor()._max_workers
        )

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor:
            results = list(
                executor.map(
                    self._invoke_helper, [(message, stop, False) for message in prompt]
                )
            )
        return results

    def invoke(
        self,
        messages: list,
        stop: Optional[List[str]] = None,
        verbose: bool = True,
        response_format: Union[dict, BaseModel, None] = None,
    ) -> str:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        return self._call(messages, stop, verbose, response_format=response_format)

    def invoke_stream(
        self, messages: list, stop: Optional[List[str]] = None
    ) -> Generator:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        return self.stream(messages, stop)

    def generate(self, prompt: str, save_path: str = "./image.png", **kwargs) -> Path:
        response = self.client.models.generate_content(
            model=self.model_name,
            contents=prompt,
            config=types.GenerateContentConfig(response_modalities=["TEXT", "IMAGE"]),
        )

        from PIL import Image
        from io import BytesIO

        path = Path(save_path)
        if path.is_dir():
            # If it's a directory, append default filename
            save_path = path / "image.png"

        for part in response.candidates[0].content.parts:
            if part.text is not None:
                print(part.text)
            elif part.inline_data is not None:
                image = Image.open(BytesIO((part.inline_data.data)))
                image.save(save_path.__str__())
                image.show()

        return path

    def edit(
        self, prompt: str, images: list[str], save_path: str = "./image.png", **kwargs
    ):
        ## process the image list
        images_source = [prompt]
        for image in images:
            if isinstance(image, str):
                # use Path to check if the image is exist
                image_path = Path(image)
                if not image_path.exists():
                    raise ValueError(f"Image {image} does not exist.")
                images_source.append(self.client.files.upload(file=image))
            elif isinstance(image, Path):
                if not image.exists():
                    raise ValueError(f"Image {image} does not exist.")
                images_source.append(self.client.files.upload(file=image.__str__()))
            else:
                raise ValueError(f"Image {image} is not a valid path or file object.")

        response = self.client.models.generate_content(
            model=self.model_name,
            contents=images_source,
            config=types.GenerateContentConfig(response_modalities=["TEXT", "IMAGE"]),
        )

        from PIL import Image
        from io import BytesIO

        path = Path(save_path)
        if path.is_dir():
            # If it's a directory, append default filename
            save_path = path / "image.png"

        for part in response.candidates[0].content.parts:
            if part.text is not None:
                print(part.text)
            elif part.inline_data is not None:
                image = Image.open(BytesIO((part.inline_data.data)))
                image.save(save_path.__str__())
                image.show()

        return path


def convert_vision_prompt(prompt: list):
    """convert the vision prompt to the correct format"""
    converted_prompts = []
    for idx, p in enumerate(prompt):
        if isinstance(p, str):
            converted_prompts.append(types.Part.from_text(text=p))
        elif isinstance(p, dict):
            if "text" in p:
                converted_prompts.append(types.Part.from_text(text=p["text"]))
            elif "image_url" in p:
                converted_prompts.append(
                    types.Part.from_bytes(data=p["image_url"], mime_type=p["mime_type"])
                )
            else:
                raise ValueError(f"Invalid prompt format: {p}")
        else:
            raise ValueError(f"Invalid prompt format: {p}")

    return converted_prompts


def check_format_prompt(prompts: list):
    """check and format the prompt to fit the correct gemini format"""
    converted_prompts = []
    system_prompt = None
    for idx, prompt in enumerate(prompts):
        if prompt["role"] == "user" or prompt["role"] == "human":
            if "parts" in prompt:
                converted_prompts.append(types.Part.from_text(text=prompt["parts"][0]))
            elif "content" in prompt:
                if isinstance(prompt["content"], list):
                    return convert_vision_prompt(prompt["content"]), system_prompt
                converted_prompts.append(types.Part.from_text(text=prompt["content"]))

        elif (
            prompt["role"] == "assistant"
            or prompt["role"] == "system"
            or prompt["role"] == "model"
        ):
            if "parts" in prompt:
                system_prompt = prompt["parts"][0]
            elif "content" in prompt:
                system_prompt = prompt["content"]

    return converted_prompts, system_prompt


def calculate_token(
    prompt: str,
    model_name: str = "gemini-1.5-flash",
):
    client = genai.Client(api_key=os.environ["GEMINI_API_KEY"])
    num_tokens = client.models.count_tokens(
        model=model_name, contents=prompt
    ).total_tokens

    return num_tokens


class gemini_embed(BaseModel, Embeddings):
    """gemini embedding models."""

    client: genai.Client = None
    model_name: str = "gemini-embedding-exp-03-07"
    """Model name to use."""
    # """Keyword arguments to pass to the model."""
    embedConfig: types.EmbedContentConfig = None
    """Keyword arguments to pass when calling the `encode` method of the model."""

    class Config:
        arbitrary_types_allowed = True  # Allow arbitrary types like genai.Client

    def __init__(self, model_name: str, api_key: str, **kwargs: Any):
        """Initialize the sentence_transformer."""
        super().__init__(**kwargs)

        self.client = genai.Client(api_key=api_key)
        self.model_name = model_name
        self.embedConfig = types.EmbedContentConfig(task_type="SEMANTIC_SIMILARITY")

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Compute doc embeddings using a HuggingFace transformer model.

        Args:
            texts: The list of texts to embed.

        Returns:
            List of embeddings, one for each text.
        """

        texts = list(map(lambda x: x.replace("\n", " "), texts))

        result = self.client.models.embed_content(
            model=self.model_name, contents=texts, config=self.embedConfig
        )
        embeddings = []
        # get list of embedding floats #
        for r in result.embeddings:
            embeddings.append(list(map(lambda x: float(x), r.values)))
        return embeddings

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a HuggingFace transformer model.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.
        """
        result = self.client.models.embed_content(
            model=self.model_name, contents=text, config=self.embedConfig
        )

        return result.embeddings[0].values



from pathlib import Path
import time
import datetime

from typing import Callable, Union, List, Generator
from langchain_core.language_models.base import BaseLanguageModel

from langchain_core.embeddings import Embeddings

from akasha.utils.base import (
    DEFAULT_CHUNK_SIZE,
    DEFAULT_EMBED,
    DEFAULT_MAX_INPUT_TOKENS,
    DEFAULT_MAX_OUTPUT_TOKENS,
    DEFAULT_MODEL,
    DEFAULT_SEARCH_TYPE,
)

from akasha.utils.atman import atman
from akasha.utils.db.db_structure import dbs

from akasha.helper.base import get_doc_length
from akasha.helper.token_counter import myTokenizer
from akasha.helper.run_llm import call_model, call_stream_model, call_batch_model
from .self_ask import self_ask_f
from akasha.helper.preprocess_prompts import merge_history_and_prompt
from akasha.utils.prompts.gen_prompt import (
    default_doc_ask_prompt,
    format_sys_prompt,
    default_get_reference_prompt,
)
from akasha.utils.search.retrievers.base import get_retrivers
from akasha.utils.search.search_doc import search_docs
import pathlib
from dotenv import load_dotenv
import warnings
import logging
from collections import defaultdict

warnings.filterwarnings("ignore", category=UserWarning, module="pydantic")
load_dotenv(pathlib.Path().cwd() / ".env", override=True)


class RAG(atman):
    """class for implement search db based on user prompt and generate response from llm model, include get_response and chain_of_thoughts."""

    def __init__(
        self,
        model: Union[str, BaseLanguageModel] = DEFAULT_MODEL,
        embeddings: Union[str, Embeddings] = DEFAULT_EMBED,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        search_type: Union[str, Callable] = DEFAULT_SEARCH_TYPE,
        max_input_tokens: int = DEFAULT_MAX_INPUT_TOKENS,
        max_output_tokens: int = DEFAULT_MAX_OUTPUT_TOKENS,
        temperature: float = 0.0,
        threshold: float = 0.0,
        language: str = "ch",
        record_exp: str = "",
        system_prompt: str = "",
        prompt_format_type: str = "auto",
        keep_logs: bool = False,
        use_chroma: bool = False,
        stream: bool = False,
        verbose: bool = False,
        env_file: str = "",
    ):
        """initials of Doc_QA class

        Args:
            embeddings (_type_, optional): embedding model, including two types(openai and huggingface). Defaults to "openai:text-embedding-ada-002".
            chunk_size (int, optional): the max length of each text segments. Defaults to 1000.
            model (_type_, optional): language model. Defaults to "openai:gpt-3.5-turbo".
            verbose (bool, optional): print the processing text or not. Defaults to False.
            threshold (float, optional): (deprecated) threshold of similarity for searching relavant documents. Defaults to 0.2.
            language (str, optional): "ch" chinese or "en" english. Defaults to "ch".
            search_type (Union[str, Callable], optional): _description_. Defaults to "auto".
            record_exp (str, optional): experiment name of aiido. Defaults to "".
            system_prompt (str, optional): the prompt you want llm to output in certain format. Defaults to "".
            prompt_format_type (str, optional): the prompt and system prompt format for the language model, including auto, gpt, llama, chat_gpt, chat_mistral, chat_gemini . Defaults to "auto".
            temperature (float, optional): temperature for language model. Defaults to 0.0.
            keep_logs (bool, optional): record logs or not. Defaults to False.
            use_chroma (bool, optional): use chroma db name instead of documents path to load data or not. Defaults to False.
            ignore_check (bool, optional): speed up loading data if the chroma db is already existed. Defaults to False.
            max_output_tokens (int, optional): max output tokens of llm model. Defaults to 1024.\n
            max_input_tokens (int, optional): max input tokens of llm model. Defaults to 3000.\n
            env_file (str, optional): the path of the .env file. Defaults to "".\n
        """

        super().__init__(
            model,
            embeddings,
            chunk_size,
            search_type,
            max_input_tokens,
            max_output_tokens,
            temperature,
            threshold,
            language,
            record_exp,
            system_prompt,
            keep_logs,
            verbose,
            use_chroma,
            env_file,
        )
        ### set argruments ###
        self.data_source = ""
        self.use_chroma = use_chroma
        self.prompt_format_type = prompt_format_type
        self.stream = stream
        ### set variables ###

        self.docs = []
        self.response = ""
        self.prompt = ""
        self.prompt_tokens, self.prompt_length = 0, 0
        self.doc_tokens, self.doc_length = 0, 0

        ## set default RAG prompt ##
        if self.system_prompt.replace(" ", "") == "":
            self.system_prompt = default_doc_ask_prompt(self.language)

    def _add_basic_log(
        self, timestamp: str, fn_type: str, history_messages: list = []
    ) -> bool:
        if super()._add_basic_log(timestamp, fn_type) is False:
            return False

        self.logs[timestamp]["prompt"] = self.prompt
        self.logs[timestamp]["history_messages"] = history_messages

        cur_source = self.data_source
        if not isinstance(cur_source, list):
            cur_source = [cur_source]

        self.logs[timestamp]["data_source"] = []

        for data_path in cur_source:
            if isinstance(data_path, Path):
                self.logs[timestamp]["data_source"].append(data_path.__str__())
            else:
                self.logs[timestamp]["data_source"].append(data_path)

        return True

    def _add_result_log(self, timestamp, time) -> bool:
        if super()._add_result_log(timestamp, time) is False:
            return False

        if self.logs[timestamp]["fn_type"] == "selfask_RAG":
            self.logs[timestamp]["follow_up"] = self.follow_up
        self.logs[timestamp]["response"] = self.response
        ### add token information ###

        self.logs[timestamp]["prompt_tokens"] = self.prompt_tokens
        self.logs[timestamp]["prompt_length"] = self.prompt_length
        self.logs[timestamp]["doc_tokens"] = self.doc_tokens
        self.logs[timestamp]["doc_length"] = self.doc_length

        return True

    def _display_info(self) -> bool:
        if self.verbose is False:
            return False
        print(f"Model: {self.model}, Embeddings: {self.embeddings}")
        print(f"Chunk size: {self.chunk_size}, Search type: {self.search_type}")
        print(
            f"Prompt tokens: {self.prompt_tokens}, Prompt length: {self.prompt_length}"
        )
        print(f"Doc tokens: {self.doc_tokens}, Doc length: {self.doc_length}\n\n")

        return True

    def _display_stream(
        self, text_input: Union[str, List[str]]
    ) -> Generator[str, None, None]:
        ret = call_stream_model(
            self.model_obj,
            text_input,
        )

        for s in ret:
            self.response += s
            yield s

    def __call__(
        self,
        data_source: Union[List[Union[str, Path]], Path, str, dbs],
        prompt: str,
        history_messages: list = [],
        **kwargs,
    ):
        """input the documents directory path and question, will first store the documents
        into vectors db (chromadb), then search similar documents based on the prompt question.
        llm model will use these documents to generate the response of the question.

            Args:
                **data_source (Union[List[Union[str, Path]], Path, str, dbs])**: documents directory path\n
                **prompt (str)**:question you want to ask.\n
                **kwargs**: the arguments you set in the initial of the class, you can change it here. Include:\n
                embeddings, chunk_size, model, verbose, topK, language , search_type, record_exp,
                system_prompt, max_doc_len, temperature.

            Returns:
                response (str): the response from llm model.
        """
        ### set variables ###

        self._set_model(**kwargs)
        self._change_variables(**kwargs)
        self.data_source = self._check_doc_path(data_source)
        self._get_db(data_source)  # create self.db and self.ignore_files
        self.prompt = prompt

        start_time = time.time()
        self._check_db()
        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        self._add_basic_log(timestamp, "RAG", history_messages)

        ### check if prompt <= max_input_tokens ###
        tot_prompts = self.prompt + self.system_prompt + "\n\n".join(history_messages)
        self.prompt_length = get_doc_length(self.language, tot_prompts)
        self.prompt_tokens = myTokenizer.compute_tokens(tot_prompts, self.model) + 10

        if self.prompt_tokens > self.max_input_tokens:
            print("\n\nThe tokens of prompt is larger than max_input_tokens.\n\n")
            raise ValueError("The tokens of prompt is larger than max_input_tokens.")

        ### start to get response ###
        retrivers_list = get_retrivers(
            self.db,
            self.embeddings_obj,
            self.threshold,
            self.search_type,
            self.env_file,
        )

        self.docs, self.doc_length, self.doc_tokens = search_docs(
            retrivers_list,
            self.prompt,
            self.model,
            self.max_input_tokens - self.prompt_tokens,
            self.search_type,
            self.language,
        )
        if self.doc_tokens == 0:
            print(
                "Warning: Unable to retrieve any documents, possibly due to insufficient remaining tokens.\n\n"
            )
            self.docs = []

        self._display_info()  # display the information of the parameters
        self._display_docs()

        text_input = merge_history_and_prompt(
            history_messages,
            self.system_prompt,
            self._format_docs() + "User question: " + self.prompt,
            self.prompt_format_type,
            model=self.model,
        )

        end_time = time.time()
        if self.stream:
            return self._display_stream(
                text_input,
            )

        self.response = call_model(
            self.model_obj,
            text_input,
        )

        self._add_result_log(timestamp, end_time - start_time)

        self._upload_logs(end_time - start_time, self.doc_length, self.doc_tokens)

        return self.response

    def reference(self) -> dict:
        """reference docs after calling the rag function, will return the reference file names of the response."""

        if self.response == "":
            logging.warning("Response empty. Please call the RAG function first.")
            print("Response empty. Please call the RAG function first.")
            return set()

        if self.docs == []:
            logging.warning("No documents found. Please call the RAG function first.")
            print("No documents found. Please call the RAG function first.")
            return set()
        ## sort out self.docs ##
        self.ref_files = defaultdict(set)
        meta_content_dict = defaultdict(str)
        for doc in self.docs:
            # doc.page_content
            if "source" in doc.metadata and "page" in doc.metadata:
                meta_content_dict[(doc.metadata["source"], doc.metadata["page"])] += (
                    doc.page_content
                )

            elif "url" in doc.metadata and "title" in doc.metadata:
                meta_content_dict[(doc.metadata["url"], doc.metadata["title"])] += (
                    doc.page_content
                )

            else:
                continue

        ## format whole prompt list ##
        prod_sys_prompts = []
        m_key_list = []

        left_tokens = self.max_input_tokens - myTokenizer.compute_tokens(
            default_get_reference_prompt()
            + "Reference: \n\n"
            + "Response: "
            + self.response,
            self.model,
        )
        for m_key, m_content in meta_content_dict.items():
            m_content_tokens = myTokenizer.compute_tokens(m_content, self.model)

            ## separate m_content if m_content_tokens larger than left_tokens
            if m_content_tokens >= left_tokens:
                pre_content = m_content[: len(m_content) // 2]
                m_content = m_content[len(m_content) // 2 :]

                pre_sys_prompt = format_sys_prompt(
                    default_get_reference_prompt(),
                    "Reference: " + pre_content + "\n\n" + "response: " + self.response,
                    self.prompt_format_type,
                    model=self.model,
                )
                prod_sys_prompts.append(pre_sys_prompt)
                m_key_list.append(m_key)

            prod_sys_prompt = format_sys_prompt(
                default_get_reference_prompt(),
                "Reference: " + m_content + "\n\n" + "response: " + self.response,
                self.prompt_format_type,
                model=self.model,
            )
            prod_sys_prompts.append(prod_sys_prompt)
            m_key_list.append(m_key)

        ## call batch model ##
        batch_responses = call_batch_model(
            self.model_obj,
            prod_sys_prompts,
        )
        print(batch_responses)  ##
        for idx, res in enumerate(batch_responses):
            res = res.lower()
            if "yes" in res:
                cur_m_key = m_key_list[idx]
                self.ref_files[cur_m_key[0]].add(cur_m_key[1])

        if self.verbose:
            print(
                "\n\nReference files: ",
                "\n".join(
                    [
                        f"{ref}: {', '.join([str(c) for c in sorted(sub)])}"
                        for ref, sub in self.ref_files.items()
                    ]
                ),
            )

        return self.ref_files

    def selfask_RAG(
        self,
        data_source: Union[List[Union[str, Path]], Path, str, dbs],
        prompt: str,
        **kwargs,
    ):
        """input the documents directory path and question, will first store the documents
        into vectors db (chromadb), then search similar documents based on the prompt question.
        question will use self-ask with search to solve complex question.
        llm model will use these documents to generate the response of the question.

            Args:
                **data_source (Union[List[Union[str, Path]], Path, str, dbs])**: documents directory path\n
                **prompt (str)**:question you want to ask.\n
                **kwargs**: the arguments you set in the initial of the class, you can change it here. Include:\n
                embeddings, chunk_size, model, verbose, language , search_type, record_exp,
                system_prompt, max_input_tokens, temperature.

            Returns:
                response (str): the response from llm model.
        """

        ### set variables ###

        self._set_model(**kwargs)
        self._change_variables(**kwargs)
        self.data_source = self._check_doc_path(data_source)
        self._get_db(data_source)  # create self.db and self.ignore_files
        self.prompt = prompt

        start_time = time.time()
        self._check_db()
        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        self._add_basic_log(timestamp, "selfask_RAG")

        ### check if prompt <= max_input_tokens ###
        tot_prompts = self.prompt + self.system_prompt
        self.prompt_length = get_doc_length(self.language, tot_prompts)
        self.prompt_tokens = myTokenizer.compute_tokens(tot_prompts, self.model) + 10

        if self.prompt_tokens > self.max_input_tokens:
            print("\n\nThe tokens of prompt is larger than max_input_tokens.\n\n")
            raise ValueError("The tokens of prompt is larger than max_input_tokens.")

        self.response = self_ask_f(self, start_time, timestamp)

        return self.response



from langchain_core.documents import Document

# from langchain_openai import OpenAIEmbeddings
from langchain.chains.query_constructor.base import AttributeInfo
from typing import List, Union, Set, Any, Tuple, Optional, Callable
import json
from pydantic import Field
import operator
from langchain_core.language_models.base import BaseLanguageModel
import akasha.helper as helper
from akasha.utils.db import dbs, extract_db_by_ids
import re
import logging


def self_query(
    prompt: str,
    model_obj: BaseLanguageModel,
    db: dbs,
    metadata_field_info: List[Union[AttributeInfo, dict]] = [],
    document_content_description: str = "",
    custom_parser: Optional[Callable[[str], Tuple[str, dict]]] = None,
    loose_filter: bool = False,
) -> Tuple[dbs, str, dict]:
    """use prompt to generate query and filter to filter out documents, return the filtered documents dbs and the query and matched fields

    Args:
        prompt (str): user question to generate query and filter
        model_obj (BaseLanguageModel): llm object
        db (dbs): the documents dbs to be filtered
        metadata_field_info (List[Union[AttributeInfo, dict]], optional): the metadata attributes info . Defaults to [].
        document_content_description (str, optional): the desciption of documents. Defaults to "".
        custom_parser (Optional[Callable[[str], Tuple[str, dict]]], optional): custom parser function. Defaults to None.

    Returns:
        Tuple[dbs, str, dict]: the filtered documents dbs, the query(str) and matched fields dict(dict[db.ids, List[str]])
    """
    query, fnl_filter, data_source = generate_query_filter(
        model_obj,
        prompt,
        metadata_field_info,
        document_content_description,
        custom_parser,
    )

    new_docs = [
        DocumentCP(doc, metadata, ids)
        for doc, metadata, ids in zip(db.docs, db.metadatas, db.ids)
    ]
    filtered_docs = filter_docs(new_docs, fnl_filter, data_source, loose_filter)
    filtered_docs_set = set(fd.ids for fd in filtered_docs)
    matched_fields_dict = {fd.ids: fd.matched_fields for fd in filtered_docs}
    new_dbs = extract_db_by_ids(db, filtered_docs_set)
    return new_dbs, query, matched_fields_dict


def check_metadata_info(
    metadata_info: List[Union[AttributeInfo, dict]],
) -> List[AttributeInfo]:
    """check if the metadata_info is in the right format"""

    new_metadata_info = []
    if isinstance(metadata_info, list):
        for i, attr in enumerate(metadata_info):
            if isinstance(attr, dict):
                if "name" in attr and "description" in attr and "type" in attr:
                    new_metadata_info.append(AttributeInfo(**attr))
                else:
                    logging.warning(
                        """The attribute info should be a dictionary with keys "name", "description" and "type"."""
                    )
            elif isinstance(attr, AttributeInfo):
                new_metadata_info.append(attr)
            else:
                logging.warning(
                    """The attribute info should be a dictionary or an instance of AttributeInfo."""
                )
    else:
        logging.warning(
            """The metadata info should be a list of dictionaries or instances of AttributeInfo."""
        )
    return new_metadata_info


def generate_query_constructor(
    metadata_info: List[Union[AttributeInfo, dict]],
    document_content_description: str,
    prompt: str,
) -> tuple[str, dict]:
    """generate a query constructor prompt based on metadata_info and document_content_description"""

    new_metadata_info = check_metadata_info(metadata_info)

    data_source = {
        "content": document_content_description,
        "attributes": {
            attr.name: {"description": attr.description, "type": attr.type}
            for attr in new_metadata_info
        },
    }
    json_string = json.dumps(data_source, ensure_ascii=False, indent=4)
    st1 = """\
<< Structured Request Schema >>
When responding use a markdown code snippet with a JSON object formatted in the following schema:

```json
{{
    "query": string \\ text string to compare to document contents
    "filter": string \\ logical condition statement for filtering documents
}}
```

The query string should contain only text that is expected to match the contents of documents. Any conditions in the filter should not be mentioned in the query as well.

A logical condition statement is composed of one or more comparison and logical operation statements.

A comparison statement takes the form: `comp(attr, val)`:
- `comp` (eq | ne | gt | gte | lt | lte): comparator
- `attr` (string):  name of attribute to apply the comparison to
- `val` (string): is the comparison value

A logical operation statement takes the form `op(statement1, statement2, ...)`:
- `op` (and | or): logical operator
- `statement1`, `statement2`, ... (comparison statements or logical operation statements): one or more statements to apply the operation to

Make sure that you only use the comparators and logical operators listed above and no others.
Make sure that filters only refer to attributes that exist in the data source.
Make sure that filters only use the attributed names with its function names if there are functions applied on them.
Make sure that filters only use format `YYYY-MM-DD` when handling date data typed values.
Make sure that filters take into account the descriptions of attributes and only make comparisons that are feasible given the type of data being stored.
Make sure that filters are only used as needed. If there are no filters that should be applied return "NO_FILTER" for the filter value.


<< Example 1. >>
Data Source:
```json
{
    "content": "Lyrics of a song",
    "attributes": {
        "artist": {
            "type": "string",
            "description": "Name of the song artist"
            },
        "length": {
            "type": "integer",
            "description": "Length of the song in seconds"
            },
        "genre": {
            "type": "string",
            "description": "The song genre, one of "pop", "rock" or "rap""
            }
        }
}
```

User Query:
What are songs by Taylor Swift or Katy Perry about teenage romance under 3 minutes long in the dance pop genre

Structured Request:
```json
{
    "query": "teenager love",
    "filter": "and(or(eq(artist, Taylor Swift), eq(artist, Katy Perry)), lt(length, 180))"
    }

```

User Query:
What are songs that were not published on Spotify

Structured Request:
```json
{    
"query": "",
"filter": "NO_FILTER"
}
```


<< Example 3. >>
Data Source:
```json
"""
    st2 = """\
```

User Query:
"""
    st3 = """\
        

Structured Request:

"""

    query_constructor_prompt = st1 + json_string + st2 + prompt + st3
    return query_constructor_prompt, data_source


class DocumentCP(Document):
    """Document subclass for comparing equality based on page content only.

    Args:
        Document (_type_): _description_

    Returns:
        _type_: _description_
    """

    ids: Union[int, str] = Field(default="")
    matched_fields: List[str] = Field(default=[])

    def __init__(self, page_content, metadata, ids):
        super().__init__(page_content)
        self.metadata = metadata
        self.ids = ids

    def __hash__(self):
        return hash(self.page_content)

    def __eq__(self, other):
        return isinstance(other, DocumentCP) and self.page_content == other.page_content


def handle_attr(
    attr: Union[int, float, str], data_source: dict, keyword: str
) -> Union[int, float, str]:
    """change attriubte type based on the data_source

    Args:
        attr (Union[int, float, str]): the attribute to be changed
        data_source (dict): the dictionary which has the data type of each attributes from documents metadata
        keyword (str): the attribute key name

    Returns:
        Union[int, float, str]:  changed type of the attribute
    """
    if data_source["attributes"][keyword]["type"] == "string":
        return str(attr)
    elif data_source["attributes"][keyword]["type"] == "integer":
        return int(attr)
    elif data_source["attributes"][keyword]["type"] == "float":
        return float(attr)
    return str(attr)


def find_subset(
    val: Union[int, float, str],
    keyword: str,
    cur_docs: List[DocumentCP],
    data_source: dict,
) -> Set[DocumentCP]:
    """base on the keyword value and operator, find the subset of documents

    Args:
        val (Union[int, float, str]): _description_
        keyword (str): _description_
        cur_docs (List[DocumentCP]): _description_
        data_source (dict): _description_

    Returns:
        Set[DocumentCP]: _description_
    """
    # Define a dictionary to map operators to functions
    operator_map = {
        "$eq": operator.eq,
        "$ne": operator.ne,
        "$gt": operator.gt,
        "$gte": operator.ge,
        "$lt": operator.lt,
        "$lte": operator.le,
    }

    res = set()
    for op, func in operator_map.items():
        if op in val:
            attr = handle_attr(val[op], data_source, keyword)

            if isinstance(attr, str) and op == "$eq":
                attr = attr.lower()
                for doc in cur_docs:
                    if (
                        keyword in doc.metadata
                        and attr in doc.metadata[keyword].lower()
                    ):
                        doc.matched_fields.append(keyword)
                        res.add(doc)
            else:
                for doc in cur_docs:
                    if keyword in doc.metadata and func(doc.metadata[keyword], attr):
                        doc.matched_fields.append(keyword)
                        res.add(doc)

    return res


def transfer_filter(filters: Union[dict, int, float, str]) -> dict:
    """Recursively convert all $and filters to $or filters."""
    if not isinstance(filters, dict):
        return filters
    elif "$and" in filters:
        filters = {"$or": [transfer_filter(f) for f in filters["$and"]]}
    else:
        for key, value in filters.items():
            filters[key] = transfer_filter(value)
    return filters


def filter_docs(
    docs: Document, filters: dict, data_source: dict, loose_filter: bool = False
) -> List[DocumentCP]:
    """filter the documents based on the filters

    Args:
        docs (Document): the documents to be filtered
        filters (dict): the filters dictionary to be applied
        data_source (dict): the data type of each attributes from documents metadata
        loose_filter (bool, optional): if True, change $and to $or. Defaults to False.

    Returns:
        List[DocumentCP]: return the filtered documents
    """
    if filters is None or len(filters) == 0:
        return docs
    if loose_filter:
        filters = transfer_filter(filters)

    def recur(cur_docs: DocumentCP, filters: dict):
        res = set()
        flag = True
        if "$and" in filters:
            for f in filters["$and"]:
                if flag:
                    flag = False
                    res = recur(cur_docs, f)
                else:
                    res = res.intersection(recur(cur_docs, f))
        elif "$or" in filters:
            for f in filters["$or"]:
                res = res.union(recur(cur_docs, f))
        else:
            keyword = next(iter(filters))
            val = filters[keyword]

            res = find_subset(val, keyword, cur_docs, data_source)

        return res

    return list(recur(docs, filters))


def translate(expr):
    """Translate a logical expression to a nested dictionary format."""
    # Pattern to match logical operations (e.g., and(...), or(...))
    logical_op_pattern = re.compile(r"(and|or|AND|OR)\((.*)\)")
    # Pattern to match comparison operations (e.g., eq(公司名稱, a公司))
    comparison_op_pattern = re.compile(r"(eq|ne|gt|gte|lt|lte)\(([^,]+),([^,]+)\)")

    # Check if it's a logical operation
    logical_match = logical_op_pattern.match(expr)
    if logical_match:
        logical_op = logical_match.group(1)  # "and" or "or"
        inner_expr = logical_match.group(2)  # Expressions inside the parentheses

        # Split inner expressions by commas not inside parentheses
        parts = split_expressions(inner_expr)
        # Recursively translate each part
        parsed_parts = [translate(part.strip()) for part in parts]
        return {f"${logical_op}": parsed_parts}

    # Check if it's a comparison operation
    comparison_match = comparison_op_pattern.match(expr)
    if comparison_match:
        comparison_op, field, value = comparison_match.groups()
        field = field.strip()
        value = value.strip()
        return {field: {f"${comparison_op}": value}}

    raise ValueError("Invalid expression format")


def split_expressions(expr):
    # Split expressions by commas outside parentheses
    parts = []
    depth = 0
    current_part = []

    for char in expr:
        if char == "," and depth == 0:
            parts.append("".join(current_part).strip())
            current_part = []
        else:
            current_part.append(char)
            if char == "(":
                depth += 1
            elif char == ")":
                depth -= 1

    if current_part:
        parts.append("".join(current_part).strip())

    return parts


def translate_output(llm_res: str) -> tuple[str, dict]:
    """translate the output of the language model to query and filter
    the output of llm may look like this:
    ```json
    {
        "query": "產品疑慮",
        "filter": "and(eq(公司名稱, a公司), eq(拜訪年, 2024))"
    }
    ```
    first we extract the query and filter from the llm_res by using helper.extract_json
    then we translate the filter to a dictionary looks like this:
    {'$and': [{'公司名稱': {'$eq': 'a公司'}}, {'拜訪年': {'$eq': '2024'}}]}

    Args:
        llm_res (str): the output of the language model

    Returns:
        tuple[str, dict]: _description_
    """
    query = ""
    filters = {}
    try:
        llm_res = helper.extract_json(llm_res)
        query = llm_res["query"]
        filter = llm_res["filter"].replace('"', "").replace("\\", "").replace("  ", "")

        filters = translate(filter)
        print(filters)
    except Exception as e:
        print(f"Error: {e}")

    return query, filters


def generate_query_filter(
    model_obj: BaseLanguageModel,
    prompt: str,
    metadata_field_info: List[Union[AttributeInfo, dict]] = [],
    document_content_description: str = "",
    custom_parser: Optional[Callable[[str], Tuple[str, dict]]] = None,
) -> tuple[str, dict, dict[str, Any]]:
    """based on metadata_field_info and document_content_description, generate a query and filter to filter out documents

    Args:
        model_obj (BaseLanguageModel): llm object
        prompt (str): the user question to generate the query and filter
        metadata_field_info (List[Union[AttributeInfo, dict]], optional): _description_. Defaults to [].
        document_content_description (str, optional): _description_. Defaults to "".
        if custom_parser is not None, it will be used to parse the output of the model
    Returns:
        tuple[str, dict, dict[str, Any]]: query, fnl_filter, data_source
    """
    query_constructor_prompt, data_source = generate_query_constructor(
        metadata_field_info, document_content_description, prompt
    )
    res = helper.call_model(model_obj, query_constructor_prompt)

    if custom_parser is not None:
        query, fnl_filter = custom_parser(res)
    else:
        query, fnl_filter = translate_output(res)

    return query, fnl_filter, data_source



from .RAG.rag import RAG
from .tools.ask import ask
from .tools.summary import summary
from .tools.websearch import websearch
from .tools.gen_img import gen_image, edit_image
from .eval import eval
from .agent import agents
from .agent import create_tool

__all__ = [
    "RAG",
    "ask",
    "summary",
    "websearch",
    "eval",
    "agents",
    "create_tool",
    "gen_image",
    "edit_image",
]



DEFAULT_MODEL = "openai:gpt-3.5-turbo"
DEFAULT_EMBED = "openai:text-embedding-ada-002"
DEFAULT_MAX_INPUT_TOKENS = 3000
DEFAULT_MAX_OUTPUT_TOKENS = 1024
DEFAULT_CHUNK_SIZE = 1000
DEFAULT_SEARCH_TYPE = "auto"



from typing import List, Optional
from langchain.llms.base import LLM

# from langchain.callbacks.manager import CallbackManager
# from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from typing import Generator, Union  # noqa: F811
from pydantic import Field
import atexit


class LlamaCPP(LLM):
    max_token: int = 4096
    model: LLM = Field(default=None)
    device: str = Field(default=None)
    model_id: str
    temperature: float = 0.01
    max_output_tokens: int = 1024
    verbose: bool = False

    def __init__(self, model_name: str, temperature: float, **kwargs):
        """define custom model, input func and temperature

        Args:
            **func (Callable)**: the function return response from llm\n
        """
        super().__init__(model_id=model_name)
        from llama_cpp import Llama

        if temperature == 0.0:
            temperature = 0.01
        if "max_output_tokens" in kwargs:
            self.max_output_tokens = kwargs["max_output_tokens"]
        if "verbose" in kwargs:
            self.verbose = kwargs["verbose"]
        self.temperature = temperature

        try:
            self.model = Llama(
                self.model_id, n_ctx=8192, n_threads=16, n_batch=512, verbose=False
            )
        except Exception:
            try:
                repo_id = "/".join(self.model_id.split("/")[:-1])
                file_name = self.model_id.split("/")[-1]
                self.model = Llama.from_pretrained(
                    repo_id=repo_id,
                    filename=file_name,
                    n_ctx=8192,
                    n_threads=16,
                    n_batch=512,
                    verbose=self.verbose,
                )
            except Exception:
                print(f"model {model_name} not found.")
                raise Exception(f"model {model_name} not found.")

        # Register the cleanup function
        atexit.register(self.cleanup)

    def cleanup(self):
        """Cleanup function to be called on exit."""
        if hasattr(self, "model") and self.model is not None:
            try:
                del self.model
            except Exception as e:
                print(f"Error during cleanup: {e}")

    @property
    def _llm_type(self) -> str:
        """return llm type

        Returns:
            str: llm type
        """
        return "llama-cpp:" + self.model_id

    def stream(
        self, prompt: Union[list, str], stop: Optional[List[str]] = None
    ) -> Generator[str, None, None]:
        stop_list = get_stop_list(stop)
        input_text = ""
        if isinstance(prompt, list):
            for pp in prompt:
                input_text += pp["content"]
        else:
            input_text = prompt
        output = self.model(
            input_text,
            stream=True,
            stop=stop_list,
            temperature=self.temperature,
            max_tokens=self.max_output_tokens,
            presence_penalty=1,
            frequency_penalty=1,
        )

        for text in output:
            delta = text["choices"][0]["text"]
            yield delta

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        stop_list = get_stop_list(stop)
        input_text = ""
        if isinstance(prompt, list):
            for pp in prompt:
                input_text += pp["content"]
        else:
            input_text = prompt

        output = self.model(
            input_text,
            stream=True,
            stop=stop_list,
            temperature=self.temperature,
            max_tokens=self.max_output_tokens,
            presence_penalty=1,
            frequency_penalty=1,
        )

        ret = ""
        for text in output:
            delta = text["choices"][0]["text"]
            print(delta, end="", flush=True)
            ret += delta

        return ret


def get_stop_list(stop: Optional[List[str]]) -> List[str]:
    """get stop list

    Args:
        stop (Optional[List[str]]): stop list

    Returns:
        List[str]: stop list
    """
    ret = ["<|eot_id|>", "<|end_header_id|>", "</s>", "<|end_of_text|>"]
    if stop is not None:
        ret = stop
    return ret



from typing import Union, List, Tuple, Callable
from langchain_core.embeddings import Embeddings
from langchain_chroma import Chroma
from chromadb.config import Settings
from langchain.docstore.document import Document
from langchain.text_splitter import RecursiveCharacterTextSplitter
from pathlib import Path
import json
import logging
import datetime
import time
import shutil
import gc
import re
from tqdm import tqdm
from akasha.helper import get_mac_address, separate_name
from akasha.helper.crawler import get_text_from_url
from akasha.helper.handle_objects import handle_embeddings_and_name
from akasha.utils.db.file_loader import load_file, get_load_file_list
from akasha.utils.db.db_structure import (
    TEXT_EXTENSIONS,
    get_storage_directory,
    FILE_LAST_CHANGE_FILE_NAME,
)
from akasha.utils.db.db_structure import (
    OLD_BUILT,
    NOT_BUILT,
    ALREADY_BUILT,
    HNSW_THRESHOLD,
)
from akasha.utils.db.delete_db import (
    delete_documents_from_chroma_by_file_name,
    delete_documents_by_directory,
)


def create_directory_db(
    directory_path: Union[str, Path],
    embeddings: Union[str, Embeddings, Callable],
    chunk_size: int,
    sleep_time: int = 60,
    env_file: str = "",
) -> Tuple[bool, list]:
    """If the documents vector storage not exist, create chromadb based on all .pdf files in doc_path.
        It will create a directory chromadb/ and save documents db in chromadb/{doc directory name}

    Args:
        directory_path (Union[str, Path]): _description_
        embeddings (Union[str, Embeddings, Callable]): _description_
        chunk_size (int): _description_
        sleep_time (int, optional): _description_. Defaults to 60.
        env_file (str, optional): _description_. Defaults to "".
        verbose (bool, optional): _description_. Defaults to False.

    Returns:
        Tuple[dbs, list]: _description_
    """
    if isinstance(directory_path, str):
        directory_path = Path(directory_path)

    if not directory_path.exists():
        logging.warning(f"{directory_path} does not exist.\n\n")
        print(f"{directory_path} does not exist.\n\n")
        return False, []

    ### get the chromadb directory name###
    embeddings, embeddings_name = handle_embeddings_and_name(
        embeddings, False, env_file
    )
    embed_type, embed_name = separate_name(embeddings_name)

    storage_directory = get_storage_directory(
        directory_path, chunk_size, embed_type, embed_name
    )
    files, db_path_names = [], []
    suc = 0
    m_time_list, suc_file_list = [], []
    for extension in TEXT_EXTENSIONS:
        files.extend(get_load_file_list(directory_path, extension))

    progress = tqdm(total=len(files), desc=f"db {directory_path.name}")
    client_settings = Settings(
        is_persistent=True,
        persist_directory=storage_directory,
        anonymized_telemetry=False,
    )
    docsearch = Chroma(
        persist_directory=storage_directory,
        embedding_function=embeddings,
        client_settings=client_settings,
        collection_metadata={"hnsw:sync_threshold": HNSW_THRESHOLD},
    )

    for file in files:
        progress.update(1)
        whole_file_path = directory_path / file
        last_m_time = whole_file_path.stat().st_mtime

        is_doc_b = _is_doc_built(storage_directory, last_m_time, file)
        if is_doc_b == ALREADY_BUILT:
            continue
        elif is_doc_b == OLD_BUILT:
            delete_documents_from_chroma_by_file_name(docsearch, file)

        file_doc = load_file(whole_file_path, file.split(".")[-1])
        if file_doc == "" or len(file_doc) == 0:
            logging.warning(f"file {file} load failed or empty.\n\n")
            print(f"file {file} load failed or empty.\n\n")

        status = create_chromadb_from_file(
            file_doc,
            docsearch,
            chunk_size,
            embeddings,
            file,
            sleep_time,
        )
        if not status:
            logging.warning(f"create chromadb for {file} failed.\n\n")
            print(f"create chromadb for {file} failed.\n\n")
            db_path_names.append(file)
        else:
            suc += 1
            suc_file_list.append(file)
            m_time_list.append(last_m_time)

    progress.close()
    id_len = len(docsearch._collection.get()["ids"])

    del docsearch
    gc.collect()

    ### delete the storage directory if no vector created ###
    if id_len == 0:
        logging.warning(f"can no create any vector from {directory_path}.\n\n")
        print(f"can no create any vector from {directory_path}.\n\n")
        shutil.rmtree(storage_directory)
        return False, db_path_names

    #### write the last modified time of the files into json file ####
    _write_docs_built_time(storage_directory, m_time_list, suc_file_list)

    return True, db_path_names


def create_single_file_db(
    file_path: Union[str, Path],
    embeddings: Union[str, Embeddings, Callable],
    chunk_size: int,
    sleep_time: int = 60,
    env_file: str = "",
) -> bool:
    """create chromadb based on a single file. It will create a directory chromadb/ and save documents db in chromadb/{doc directory name}

    Args:
        file_path (Union[str, Path]): _description_
        embeddings (Union[str, Embeddings]): _description_
        chunk_size (int): _description_
        sleep_time (int, optional): _description_. Defaults to 60.
        env_file (str, optional): _description_. Defaults to "".

    Returns:
        bool: _description_
    """
    try:
        if isinstance(file_path, str):
            file_path = Path(file_path)
        doc_path = file_path.parent
        file_name = file_path.name
    except Exception:
        logging.warning("file path error.\n\n")
        print("file path error.\n\n")
        return False

    ### get the chromadb directory name###
    embeddings, embeddings_name = handle_embeddings_and_name(
        embeddings, False, env_file
    )
    embed_type, embed_name = separate_name(embeddings_name)
    storage_directory = get_storage_directory(
        doc_path, chunk_size, embed_type, embed_name
    )

    last_m_time = file_path.stat().st_mtime
    loaded = False
    is_doc_b = _is_doc_built(storage_directory, last_m_time, file_name)
    if is_doc_b == ALREADY_BUILT:
        return True
    elif is_doc_b == OLD_BUILT:
        client_settings = Settings(
            is_persistent=True,
            persist_directory=storage_directory,
            anonymized_telemetry=False,
        )
        docsearch = Chroma(
            persist_directory=storage_directory,
            embedding_function=embeddings,
            client_settings=client_settings,
            collection_metadata={"hnsw:sync_threshold": HNSW_THRESHOLD},
        )
        loaded = True
        delete_documents_from_chroma_by_file_name(docsearch, file_name)

    file_doc = load_file(doc_path / file_name, file_name.split(".")[-1])
    if file_doc == "" or len(file_doc) == 0:
        logging.warning("file load failed or empty.\n\n")
        print("file load failed or empty.\n\n")
        return False

    if not loaded:
        client_settings = Settings(
            is_persistent=True,
            persist_directory=storage_directory,
            anonymized_telemetry=False,
        )
        docsearch = Chroma(
            persist_directory=storage_directory,
            embedding_function=embeddings,
            client_settings=client_settings,
            collection_metadata={"hnsw:sync_threshold": HNSW_THRESHOLD},
        )

    status = create_chromadb_from_file(
        file_doc,
        docsearch,
        chunk_size,
        embeddings,
        file_name,
        sleep_time,
    )

    del docsearch
    gc.collect()

    if not status:
        logging.warning(f"create chromadb for {file_name} failed.\n\n")
        print(f"create chromadb for {file_name} failed.\n\n")

        return False

    _write_docs_built_time(storage_directory, [last_m_time], [file_name])

    return True


def create_webpage_db(
    url: str,
    embeddings: Union[str, Embeddings, Callable],
    chunk_size: int,
    sleep_time: int = 60,
    env_file: str = "",
) -> bool:
    """create chromadb from a webpage. It will create a directory chromadb/ and save documents db in chromadb/{doc directory name}

    Args:
        url (str): _description_
        embeddings (Union[str, Embeddings, Callable]): _description_
        chunk_size (int): _description_
        sleep_time (int, optional): _description_. Defaults to 60.
        env_file (str, optional): _description_. Defaults to "".

    Returns:
        bool: _description_
    """

    if not isinstance(url, str):
        logging.warning("url error.\n\n")
        print("url error, url must be string.\n\n")
        return False

    web_link_pattern = re.compile(r"^(http|https)://")
    if not web_link_pattern.match(url):
        logging.warning("url error.\n\n")
        print("url error, url must be a valid url.\n\n")
        return False

    ### get the chromadb directory name###
    embeddings, embeddings_name = handle_embeddings_and_name(
        embeddings, False, env_file
    )
    embed_type, embed_name = separate_name(embeddings_name)
    storage_directory = get_storage_directory(url, chunk_size, embed_type, embed_name)

    cur_time = datetime.datetime.now()
    last_m_time = cur_time.strftime("%Y-%m-%d %H:%M:%S:%f")
    is_doc_b = _is_url_built(storage_directory, cur_time, url)
    if is_doc_b == ALREADY_BUILT:
        return True
    elif is_doc_b == OLD_BUILT:
        delete_documents_by_directory(url, embeddings, chunk_size)

    file_title, file_doc = get_text_from_url(url)
    if file_doc == "" or len(file_doc) == 0:
        logging.warning("file load failed or empty.\n\n")
        print("file load failed or empty.\n\n")
        return False
    file_doc = [
        Document(page_content=file_doc, metadata={"title": file_title, "url": url})
    ]

    client_settings = Settings(
        is_persistent=True,
        persist_directory=storage_directory,
        anonymized_telemetry=False,
    )
    docsearch = Chroma(
        persist_directory=storage_directory,
        embedding_function=embeddings,
        client_settings=client_settings,
        collection_metadata={"hnsw:sync_threshold": HNSW_THRESHOLD},
    )

    status = create_chromadb_from_file(
        file_doc,
        docsearch,
        chunk_size,
        embeddings,
        url,
        sleep_time,
    )

    del docsearch
    gc.collect()

    if not status:
        logging.warning(f"create chromadb for {url} failed.\n\n")
        print(f"create chromadb for {url} failed.\n\n")

        return False

    _write_docs_built_time(storage_directory, [last_m_time], [url])

    return True


def create_chromadb_from_file(
    documents: List[Document],
    docsearch: Chroma,
    chunk_size: int,
    embeddings: Embeddings,
    file_name: str,
    sleep_time: int = 60,
) -> bool:
    """load the existing chromadb of documents from storage_directory and return it, if not exist, create it.

    Args:
        documents (list): list of Documents
        docsearch (Chroma): chromadb object
        chunk_size (int): the chunk size of documents
        embeddings (vars): the embeddings used in transfer documents into vector storage
        file_name (str): the path and name of the file
        sleep_time (int, optional): waiting time if exceed api calls. Defaults to 60.

    Returns:
        (chromadb object): return the dbs object
    """
    text_splitter = RecursiveCharacterTextSplitter(
        separators=["\n", " ", ",", ".", "。", "!"],
        chunk_size=chunk_size,
        chunk_overlap=100,
    )
    k = 0
    cum_ids = 0
    interval = 10
    mac_address = get_mac_address()
    is_added = False
    while k < len(documents):
        cur_doc = documents[k : k + interval]
        texts = text_splitter.split_documents(cur_doc)
        formatted_date = datetime.datetime.now().strftime("%Y-%m-%d-%H_%M_%S_%f")
        page_contents = []
        text_str_count = 0
        for text in texts:
            text_str_count += len(text.page_content)
            page_contents.append(text.page_content)

        try:
            vectors = embeddings.embed_documents(page_contents)
        except Exception:
            time.sleep(sleep_time)
            vectors = embeddings.embed_documents(page_contents)

        if len(vectors) == 0:
            k += interval
            cum_ids += len(texts)
            continue
        docsearch._collection.add(
            embeddings=vectors,
            metadatas=[text.metadata for text in texts],
            documents=[text.page_content for text in texts],
            ids=[
                formatted_date + "_" + str(cum_ids + i) + "_" + mac_address
                for i in range(len(texts))
            ],
        )

        if text_str_count > 0:
            is_added = True

        k += interval
        cum_ids += len(texts)

    if not is_added:
        logging.warning(f"\n{file_name} has empty content, ignored.\n")
        print(f"\n{file_name} has empty content, ignored.\n")
        return False
    return True


def _is_doc_built(storage_directory: str, last_m_time: float, file_name: str) -> str:
    storage_path = Path(storage_directory)
    if not storage_path.exists():
        return NOT_BUILT

    # Path to the JSON file
    json_file_path = storage_path / FILE_LAST_CHANGE_FILE_NAME

    # Check if the JSON file exists
    if not json_file_path.exists():
        return NOT_BUILT

    # Load the JSON file
    try:
        with open(json_file_path, "r", encoding="utf-8") as json_file:
            file_last_changed = json.load(json_file)
    except Exception as e:
        logging.warning(f"Error reading JSON file: {e}")
        print(f"Error reading JSON file: {e}")
        return NOT_BUILT

    # Check if the file_name is in the dictionary
    if file_name not in file_last_changed:
        return NOT_BUILT

    # Check if the last modify time matches
    if file_last_changed[file_name] != last_m_time:
        return OLD_BUILT

    return ALREADY_BUILT


def _is_url_built(
    storage_directory: str, cur_time: datetime.datetime, file_name: str
) -> str:
    storage_path = Path(storage_directory)
    if not storage_path.exists():
        return NOT_BUILT

    # Path to the JSON file
    json_file_path = storage_path / FILE_LAST_CHANGE_FILE_NAME

    # Check if the JSON file exists
    if not json_file_path.exists():
        return NOT_BUILT

    # Load the JSON file
    try:
        with open(json_file_path, "r", encoding="utf-8") as json_file:
            file_last_changed = json.load(json_file)
    except Exception as e:
        logging.warning(f"Error reading JSON file: {e}")
        print(f"Error reading JSON file: {e}")
        return NOT_BUILT

    # Check if the file_name is in the dictionary
    if file_name not in file_last_changed:
        return NOT_BUILT

    # Check if the last modify time matches
    import os

    reload_url_days = int(os.getenv("RELOAD_URL_DAYS", 30))
    file_last_changed_time = datetime.datetime.strptime(
        file_last_changed[file_name], "%Y-%m-%d %H:%M:%S:%f"
    )
    threshold_date = file_last_changed_time + datetime.timedelta(days=reload_url_days)

    if threshold_date < cur_time:
        return OLD_BUILT

    return ALREADY_BUILT


def _write_docs_built_time(
    storage_directory: str, last_m_time: List[Union[float, str]], file_name: List[str]
):
    storage_path = Path(storage_directory)

    # Path to the JSON file
    json_file_path = storage_path / FILE_LAST_CHANGE_FILE_NAME

    # Create the dictionary
    file_last_changed = {}

    # Check if the JSON file exists
    if json_file_path.exists():
        # Load the JSON file
        try:
            with open(json_file_path, "r", encoding="utf-8") as json_file:
                file_last_changed = json.load(json_file)
        except Exception as e:
            logging.warning(f"Error reading JSON file: {e}")
            print(f"Error reading JSON file: {e}")

    # Update the dictionary
    for f_name, m_time in zip(file_name, last_m_time):
        file_last_changed[f_name] = m_time

    # Write the dictionary to the JSON file
    try:
        with open(json_file_path, "w", encoding="utf-8") as json_file:
            json.dump(file_last_changed, json_file, indent=4, ensure_ascii=False)
    except Exception as e:
        logging.warning(f"Error writing JSON file: {e}")
        print(f"Error writing JSON file: {e}")



import logging
from typing import List, Any, Optional, Generator
from langchain.llms.base import LLM
from openai import OpenAI
import concurrent.futures

from pydantic import Field


class remote_model(LLM):
    max_token: int = 4096
    max_output_tokens: int = 1024
    temperature: float = 0.01
    top_p: float = 0.95
    history: list = []
    model_name: str = "remote_model"
    api_key: str = "123"
    tokenizer: Any = Field(default=None)
    model: OpenAI = Field(default=None)
    url: Any = Field(default=None)

    def __init__(
        self,
        base_url: str,
        temperature: float = 0.001,
        api_key: str = "123",
        model_name: str = "remote_model",
        **kwargs,
    ):
        """define custom model, input func and temperature

        Args:
            **func (Callable)**: the function return response from llm\n
        """
        super().__init__()
        self.url = handle_url(base_url)
        self.temperature = temperature
        self.api_key = api_key
        self.model_name = model_name
        if "max_output_tokens" in kwargs:
            self.max_output_tokens = kwargs["max_output_tokens"]

        if self.temperature == 0.0:
            self.temperature = 0.01
        self.model = OpenAI(base_url=self.url, api_key=self.api_key)

    @property
    def _llm_type(self) -> str:
        """return llm type

        Returns:
            str: llm type
        """
        return "remote: api model"

    def stream(self, prompt: str, stop: Optional[List[str]] = None) -> Generator:
        """run llm and get the stream generator

        Args:
            prompt (str): _description_

        Yields:
            Generator: _description_
        """
        if isinstance(prompt, str):
            prompt = [{"role": "user", "content": prompt}]

        yield from self.invoke_stream(prompt, stop)
        return

    def _call(self, prompt: str, stop: Optional[List[str]] = None, verbose=True) -> str:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """

        if isinstance(prompt, str):
            prompt = [{"role": "user", "content": prompt}]

        return self.invoke(prompt, stop, verbose)

    def _invoke_helper(self, args):
        messages, stop, verbose = args
        return self._call(messages, stop, verbose)

    def batch(self, prompt: List[str], stop: Optional[List[str]] = None) -> List[str]:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        # Number of threads should not exceed the number of prompts
        num_threads = min(
            len(prompt), concurrent.futures.thread.ThreadPoolExecutor()._max_workers
        )

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor:
            results = list(
                executor.map(
                    self._invoke_helper, [(message, stop, False) for message in prompt]
                )
            )
        return results

    def invoke(
        self, messages: list, stop: Optional[List[str]] = None, verbose: bool = True
    ) -> str:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        stop_list = get_stop_list(stop)

        response = ""
        try:
            chat_completion = self.model.chat.completions.create(
                messages=messages,
                model=self.model_name,
                stream=True,
                max_tokens=self.max_output_tokens,
                temperature=self.temperature,
                top_p=0.95,
                stop=stop_list,
                frequency_penalty=1.2,
            )

            for message in chat_completion:
                content = message.choices[0].delta.content
                if isinstance(content, str):
                    if verbose:
                        print(message.choices[0].delta.content, end="")
                    response += message.choices[0].delta.content

        except Exception as e:
            logging.error("call remote model failed\n\n", e.__str__())
            raise e
        return response

    def invoke_stream(
        self, messages: list, stop: Optional[List[str]] = None
    ) -> Generator:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        stop_list = get_stop_list(stop)
        url = self.url
        if url[-1] != "/":
            url += "/"

        if url[-3:] != "v1/":
            url = url + "v1/"

        try:
            chat_completion = self.model.chat.completions.create(
                messages=messages,
                model=self.model_name,
                stream=True,
                max_tokens=self.max_output_tokens,
                temperature=self.temperature,
                top_p=0.95,
                stop=stop_list,
                frequency_penalty=1.2,
            )

            for message in chat_completion:
                content = message.choices[0].delta.content
                if isinstance(content, str):
                    yield message.choices[0].delta.content

        except Exception as e:
            info = "call remote model failed\n\n"
            logging.error(info, e.__str__())
            yield info + e.__str__()


def get_stop_list(stop: Optional[List[str]]) -> List[str]:
    """get stop list

    Args:
        stop (Optional[List[str]]): stop list

    Returns:
        List[str]: stop list
    """
    ret = ["<|eot_id|>", "<|end_header_id|>", "</s>"]
    if stop is not None:
        ret = stop
    return ret


def handle_url(url: str):
    if url[-1] != "/":
        url += "/"

    if url[-3:] != "v1/":
        url = url + "v1/"
    return url



from typing import Any, List, Optional, Tuple
from pydantic import Field
import numpy as np

from langchain.schema import BaseRetriever, Document

import jieba
from rank_bm25 import BM25Okapi


class myBM25Retriever(BaseRetriever):
    bm25: BM25Okapi = Field(default=None)
    """BM25 class to use."""
    texts: List[str] = Field(default=None)
    """List of texts to index."""
    metadata: List[dict] = Field(default=None)
    docs: List[Document] = Field(default=None)
    k: int = 3
    """Number of results to return."""
    relevancy_threshold: Optional[float] = None

    @classmethod
    def from_documents(
        cls,
        docs: List[Document],
        k: int = 3,
        relevancy_threshold: float = 0.0,
        **kwargs: Any,
    ) -> BaseRetriever:
        tokenize_corpus = [list(jieba.cut(doc.page_content)) for doc in docs]
        bm25 = BM25Okapi(tokenize_corpus)
        return cls(
            bm25=bm25,
            texts=[doc.page_content for doc in docs],
            metadata=[doc.metadata for doc in docs],
            k=k,
            docs=docs,
            relevancy_threshold=relevancy_threshold,
            **kwargs,
        )

    def _gs(self, query: str) -> Tuple[List[Document], List[float]]:
        """implement bm25 to find relevant documents

        Args:
            **query (str)**: query string that used to find relevant documents\n

        Returns:
            List[Document]: relevant documents
        """

        tokenize_query = list(jieba.cut(query))
        docs_scores = self.bm25.get_scores(tokenize_query)
        top_k_idx = np.argsort(docs_scores)[::-1][: self.k]
        top_k_results = [self.docs[i] for i in top_k_idx]
        top_k_scores = [docs_scores[i] for i in top_k_idx]
        return top_k_results, top_k_scores

    def get_relevant_documents_and_scores(
        self,
        query: str,
    ) -> Tuple[List[Document], List[float]]:
        return self._gs(query)

    async def _aget_relevant_documents(self, query: str) -> List[Document]:
        return self._gs(query)[0]

    def _get_relevant_documents(self, query: str) -> List[Document]:
        return self._gs(query)[0]



from typing import List, Tuple, Union, Callable
from akasha.helper.base import get_doc_length
from akasha.helper.token_counter import myTokenizer
from .auto_search import get_relevant_doc_auto, get_relevant_doc_auto_rerank
from langchain.schema import Document, BaseRetriever
from langchain_core.language_models.base import BaseLanguageModel


def _merge_docs(
    docs_list: List[List[Document]],
    topK: int,
    language: str,
    max_input_tokens: int = 3000,
    model: str = "openai:gpt-3.5-turbo",
) -> Tuple[list, int]:
    """merge different search types documents, if total len of documents too large,
        will not select all documents.
        use jieba to count length of chinese words, use split space otherwise.

    Args:
        **docs_list (list)**: list of all docs from selected search types\n
        **topK (int)**: for each search type, select topK documents\n
        **language (str)**: 'ch' for chinese, otherwise use split space to count words, default is chinese\n
        **verbose (bool)**: show log texts or not. Defaults to False.\n
        **max_token (int)**: max token size of llm input.\n

    Returns:
        list: merged list of Documents
    """
    res = []
    cur_count, cur_token = 0, 0
    page_contents = set()
    for i in range(topK):
        for docs in docs_list:
            if i >= len(docs):
                continue

            if docs[i].page_content in page_contents:
                continue

            words_len = get_doc_length(language, docs[i].page_content)
            # token_len = model.get_num_tokens(docs[i].page_content)
            token_len = myTokenizer.compute_tokens(docs[i].page_content, model)
            if cur_token + token_len > max_input_tokens:
                return res, cur_count, cur_token

            cur_count += words_len
            cur_token += token_len
            res.append(docs[i])
            page_contents.add(docs[i].page_content)

    return res, cur_count, cur_token


def search_docs(
    retriver_list: list,
    query: str,
    model: Union[str, BaseLanguageModel] = "openai:gpt-3.5-turbo",
    max_input_tokens: int = 3000,
    search_type: Union[str, Callable] = "auto",
    language: str = "ch",
) -> Tuple[list, int, int]:
    """search docs based on given search_type, default is merge, which contain 'mmr', 'svm', 'tfidf'
        and merge them together.

    Args:
        **retriver_list (list)**: list of retrievers that the search_type needed\n
        **query (str)**: the query str used to search similar documents\n
        **language (str)**: default to chinese 'ch', otherwise english, the language of documents and prompt,
            use to make sure docs won't exceed max token size of llm input.\n
        **search_type (str)**: search type to find similar documents from db, default 'merge'.
            includes 'merge', 'mmr', 'svm', 'tfidf'.\n
        **model ()**: large language model name\n
        **max_input_tokens (int)**: max token size of llm input.\n

    Returns:
        list: selected list of similar documents.
    """

    topK = 10000

    final_docs = []
    if isinstance(model, BaseLanguageModel):
        try:
            model = model._llm_type
        except Exception:
            model = "openai:gpt-3.5-turbo"

    if not callable(search_type):
        search_type = search_type.lower()

        if search_type == "auto":
            docs = get_relevant_doc_auto(
                retriver_list,
                query,
            )
            docs, docs_len, tokens = _merge_docs(
                [docs], topK, language, max_input_tokens, model
            )
            return docs, docs_len, tokens
        elif search_type == "auto_rerank":
            docs = get_relevant_doc_auto_rerank(
                retriver_list,
                query,
                topK,
            )
            docs, docs_len, tokens = _merge_docs(
                [docs], topK, language, max_input_tokens, model
            )
            return docs, docs_len, tokens

    for retri in retriver_list:
        docs = retri._get_relevant_documents(query)
        final_docs.append(docs)

    docs, docs_len, tokens = _merge_docs(
        final_docs, topK, language, max_input_tokens, model
    )

    return docs, docs_len, tokens


def retri_docs(
    retriver_list: List[BaseRetriever],
    query: str,
    search_type: Union[str, Callable],
    topK: int,
) -> list:
    """search docs based on given search_type, default is merge, which contain 'mmr', 'svm', 'tfidf'
        and merge them together.

    Args:
        **retriver_list (list)**: list of retrievers that the search_type needed\n
        **query (str)**: the query str used to search similar documents\n
        **topK (int)**: for each search type, return first topK documents\n
        **search_type (str)**: search type to find similar documents from db, default 'merge'.
            includes 'merge', 'mmr', 'svm', 'tfidf'.\n

    Returns:
        list: selected list of similar documents.
    """

    final_docs = []

    def merge_docs(
        docs_list: List[Document],
        topK: int,
    ):
        res = []
        page_contents = set()

        for i, adocs in enumerate(docs_list):
            if len(page_contents) >= topK:
                break

            if adocs.page_content in page_contents:
                continue
            res.append(adocs)
            page_contents.add(adocs.page_content)
        return res

    if not callable(search_type):
        search_type = search_type.lower()

        if search_type == "auto":
            docs = get_relevant_doc_auto(
                retriver_list,
                query,
            )
            docs = merge_docs(docs, topK)
            return docs
        elif search_type == "auto_rerank":
            docs = get_relevant_doc_auto_rerank(
                retriver_list,
                query,
                topK,
            )
            docs = merge_docs(docs, topK)
            return docs

    for retri in retriver_list:
        docs = retri._get_relevant_documents(query)
        # docs, scores = retri._gs(query)
        final_docs.extend(docs)

    docs = merge_docs(final_docs, topK)

    return docs



from typing import Union, List, Generator
from pydantic import BaseModel
from langchain_core.messages.ai import AIMessage
import traceback
import logging
from langchain_core.language_models.base import BaseLanguageModel
from akasha.helper.handle_objects import handle_model_and_name
from akasha.helper.base import sim_to_trad, extract_json, extract_multiple_json
from akasha.utils.prompts.gen_prompt import format_sys_prompt, default_translate_prompt


def call_model(
    model: BaseLanguageModel,
    input_text: Union[str, list],
) -> str:
    """call llm model and return the response

    Args:
        model (BaseLanguageModel): llm model
        input_text (str): the input_text that send to llm model
        prompt_type (str): the type of prompt, default "gpt"

    Returns:
        str: llm response
    """

    ### for openai, change system prompt and prompt into system meg and human meg ###

    response = ""
    print_flag = True

    model, model_name = handle_model_and_name(model)

    try:
        try:
            model_type = model._llm_type
        except Exception:
            print_flag = False
            model_type = "unknown"

        if "openai" in model_type:
            print_flag = False
            response = model.invoke(input_text)

        elif "remote" in model_type:
            print_flag = False
            response = model._call(input_text)
        else:
            response = model._call(input_text)

        if isinstance(response, AIMessage):
            response = response.content
            if isinstance(response, dict):
                response = response.__str__()
            if isinstance(response, list):
                response = "\n".join(response)

        if (
            ("huggingface" in model_type)
            or ("llama cpp" in model_type)
            or ("gemini" in model_type)
            or ("anthropic" in model_type)
        ):
            print_flag = False

        if response is None or response == "":
            print_flag = False
            raise Exception("LLM response is empty.")

    except Exception as e:
        trace_text = traceback.format_exc()
        logging.error(
            trace_text
            + "\n\nText generation encountered an error.\
            Please check your model setting.\n\n"
        )
        raise e

    if print_flag:
        print("llm response:", "\n\n" + response)

    if isinstance(response, str):
        response = sim_to_trad(response)

    return response


def call_batch_model(
    model: BaseLanguageModel,
    input_text: list,
) -> List[str]:
    """call llm model in batch and return the response

    Args:
        model (BaseLanguageModel): llm model
        input_text: list

    Returns:
        str: llm response
    """

    ### check the input prompt and system prompt ###
    if isinstance(input_text, str):
        input_text = [input_text]

    response = ""
    responses = []
    model, model_name = handle_model_and_name(model)

    try:
        response = model.batch(input_text)
        for res in response:
            if isinstance(res, AIMessage):
                res = res.content
            if isinstance(res, dict):
                res = res.__str__()
            if isinstance(res, list):
                res = "\n".join(res)
            responses.append(res)

        if response is None or response == "" or "".join(responses) == "":
            # print_flag = False
            raise Exception("LLM response is empty.")

    except Exception as e:
        trace_text = traceback.format_exc()
        logging.error(
            trace_text
            + "\n\nText generation encountered an error.\
            Please check your model setting.\n\n"
        )
        raise e

    # if print_flag:
    #     print("llm response:", "\n\n" + response)

    return responses


def call_stream_model(
    model: BaseLanguageModel,
    input_text: Union[str, list],
) -> Generator[str, None, None]:
    """call llm model and yield the response

    Args:
        model (BaseLanguageModel): llm model
        input_text (str): the input_text that send to llm model
        prompt_type (str): the type of prompt, default "gpt"

    Returns:
        str: llm response
    """

    ### for openai, change system prompt and prompt into system meg and human meg ###

    response = None
    texts = ""
    model, model_name = handle_model_and_name(model)
    try:
        try:
            response = model.stream(input_text)
        except Exception:
            response = model._call(input_text)

        for r in response:
            if isinstance(r, AIMessage):
                r = r.content
                if isinstance(r, dict):
                    r = r.__str__()
                if isinstance(r, list):
                    r = "\n".join(r)
            texts += r
            yield sim_to_trad(r)

        if texts == "":
            yield "ERROR! LLM response is empty.\n\n"

    except Exception as e:
        trace_text = traceback.format_exc()
        logging.error(
            trace_text
            + "\n\nText generation encountered an error.\
            Please check your model setting.\n\n"
        )
        yield e


def call_image_model(
    model: BaseLanguageModel,
    input_text: Union[str, list],
) -> str:
    response = ""
    print_flag = True
    model, model_name = handle_model_and_name(model)
    try:
        model_type = model._llm_type

        if (
            ("openai" in model_type)
            or ("remote" in model_type)
            or ("gemini" in model_type)
        ):
            print_flag = False
            response = model.invoke(input_text)

        else:
            response = model.call_image(input_text)

        if isinstance(response, AIMessage):
            response = response.content
            if isinstance(response, dict):
                response = response.__str__()
            if isinstance(response, list):
                response = "\n".join(response)

        if response is None or response == "":
            print_flag = False
            raise Exception("LLM response is empty.")

    except Exception as e:
        trace_text = traceback.format_exc()
        logging.error(
            trace_text
            + "\n\nText generation encountered an error.\
            Please check your model setting.\n\n"
        )
        raise e

    response = sim_to_trad(response)

    if print_flag:
        print("llm response:", "\n\n" + response)
    return response


def check_relevant_answer(
    model_obj: BaseLanguageModel,
    batch_responses: List[str],
    question: str,
    prompt_format_type: str = "auto",
) -> List[str]:
    """ask LLM that each of the retrieved answers list is relevant to the question or not"""
    from akasha.utils.prompts.gen_prompt import (
        default_answer_grader_prompt,
        format_sys_prompt,
    )

    results = []
    txts = []
    sys_prompt = default_answer_grader_prompt()
    model_obj, model_name = handle_model_and_name(model_obj)
    for idx in range(len(batch_responses)):
        prod_prompt = f"Retrieved answer: \n\n {batch_responses[idx]} \n\n User question: {question}"
        text_input = format_sys_prompt(
            sys_prompt, prod_prompt, prompt_format_type, model_name
        )
        txts.append(text_input)

    response_list = call_batch_model(model_obj, txts)
    for idx, response in enumerate(response_list):
        if "yes" in response.lower():
            results.append(batch_responses[idx])

    return results


def call_translator(
    model_obj: BaseLanguageModel,
    texts: str,
    prompt_format_type: str = "auto",
    language: str = "zh",
) -> str:
    """translate texts to target language

    Args:
        model_obj (BaseLanguageModel): LLM that used to translate
        texts (str): texts that need to be translated
        prompt_format_type (str, optional): system prompt format. Defaults to "auto".
        language (str, optional): target language. Defaults to "zh".

    Returns:
        str: translated texts
    """
    model_obj, model_name = handle_model_and_name(model_obj)
    sys_prompt = default_translate_prompt(language)
    prod_prompt = format_sys_prompt(sys_prompt, texts, prompt_format_type, model_name)

    response = call_model(model_obj, prod_prompt)

    return response


def call_JSON_formatter(
    model_obj: BaseLanguageModel,
    texts: str,
    keys: Union[str, list, BaseModel] = "",
    prompt_format_type: str = "auto",
) -> Union[dict, List[dict], None]:
    """use LLM to transfer texts into JSON format

    Args:
        model_obj (BaseLanguageModel): LLM that used to transfer
        texts (str): texts that need to be transferred
        keys (Union[str, list], optional): keys name of output dictionary. Defaults to "".
        prompt_format_type (str, optional): system prompt format. Defaults to "auto".

    Returns:
        Union[dict, None]: return the JSON part of the string, if not found return None
    """

    ### RESPONSE FORMAT FAILED OR NOT OPENAI/GEMINI MODEL ###
    if keys == "":
        sys_prompt = "Format the following TEXTS into a single JSON instance that conforms to the JSON schema."
    elif isinstance(keys, str):
        keys = [keys]

    model_obj, model_name = handle_model_and_name(model_obj)
    model_name = model_name.lower()

    if keys != "":
        if not isinstance(keys, list):
            keys_list = basemodel_keys_list(keys)
        else:
            keys_list = keys

        sys_prompt = f"Format the following TEXTS into a single JSON instance that conforms to the JSON schema which includes: {', '.join(keys_list)}\n\n"

    prod_prompt = format_sys_prompt(
        sys_prompt, "TEXTS: " + texts, prompt_format_type, model_name
    )

    ## try use response format ##
    if keys != "" and ("openai" in model_name or "gemini" in model_name):
        try:
            if issubclass(keys, BaseModel):
                json_base_model = keys
            else:
                json_base_model = keys_to_basemodel_class(keys)

            if "openai" in model_name:
                response_format = {"type": "json_object"}

                # can not use it yet #
                # else:
                #    response_format = json_base_model
                response = model_obj.invoke(
                    prod_prompt,
                    response_format=response_format,
                )
                return extract_json(response)

            else:
                response = model_obj.invoke(
                    prod_prompt,
                    response_format=json_base_model,
                )
                return extract_multiple_json(response)
        except Exception as e:
            print("Error in using JSON response format:", e)

    response = call_model(model_obj, prod_prompt)
    return extract_json(response)


def keys_to_basemodel_class(
    keys: list[str], class_name: str = "JSONFormatModel"
) -> type:
    """
    Dynamically create a Pydantic BaseModel class with all fields as str.

    Args:
        keys (list[str]): List of field names.
        class_name (str): Name of the generated class.

    Returns:
        type: A new BaseModel subclass.
    """
    return type(class_name, (BaseModel,), {"__annotations__": {k: str for k in keys}})


def basemodel_keys_list(model: type) -> list[str]:
    """
    Return all field names of a Pydantic BaseModel class as a list of strings.

    Args:
        model (type): A Pydantic BaseModel class.

    Returns:
        list[str]: List of field names.
    """
    return list(model.__annotations__.keys())



from typing import Any, List, Optional, Tuple
from pydantic import Field
import numpy as np
from langchain.embeddings.base import Embeddings
from langchain.schema import BaseRetriever, Document
from akasha.utils.db.db_structure import dbs
from langchain_community.retrievers import KNNRetriever
import faiss


class myFAISSRetriever(BaseRetriever):
    embeddings: Embeddings = Field(default=None)
    """Embeddings model to use."""
    index: faiss.IndexFlatL2 = Field(default=None)
    """Index of embeddings."""
    texts: List[str] = Field(default=None)
    """List of texts to index."""
    metadata: List[dict]
    k: int = 10
    """Number of results to return."""
    relevancy_threshold: Optional[float] = None

    @classmethod
    def from_db(
        cls,
        db: dbs,
        embeddings: Embeddings,
        k: int = 10,
        relevancy_threshold: float = 0.0,
        **kwargs: Any,
    ) -> KNNRetriever:
        emb = np.array(db.get_embeds()).astype("float32")
        dimention = emb.shape[1]
        index = faiss.IndexFlatL2(dimention)
        index.add(emb)  # add vectors to the index
        texts = db.get_docs()
        metadata = db.get_metadatas()
        return cls(
            embeddings=embeddings,
            index=index,
            texts=texts,
            metadata=metadata,
            k=k,
            relevancy_threshold=relevancy_threshold,
            **kwargs,
        )

    def _gs(self, query: str) -> Tuple[List[Document], List[float]]:
        """implement FAISS search to find relevant documents

        Args:
            **query (str)**: query string that used to find relevant documents\n

        Returns:
            List[Document]: relevant documents
        """
        query_embeds = np.array([self.embeddings.embed_query(query)]).astype("float32")
        # use index to search
        distances, indexes = self.index.search(
            query_embeds, min(self.k, len(self.texts))
        )

        top_k_scores = 1 / (1 + distances[0])
        top_k_results = [
            Document(page_content=self.texts[row], metadata=self.metadata[row])
            for row in indexes[0]
            if (
                self.relevancy_threshold is None
                or top_k_scores[row] >= self.relevancy_threshold
            )
        ]
        return top_k_results, top_k_scores

    def get_relevant_documents_and_scores(
        self,
        query: str,
    ) -> Tuple[List[Document], List[float]]:
        return self._gs(query)

    def _aget_relevant_documents(self, query: str) -> List[Document]:
        """implement FAISS search to find relevant documents

        Args:
            **query (str)**: query string that used to find relevant documents\n

        Returns:
            List[Document]: relevant documents
        """
        return self._gs(query)[0]

    def _get_relevant_documents(self, query: str) -> List[Document]:
        return self._gs(query)[0]



from typing import Callable, Union, List
from akasha.helper import handle_embeddings, handle_model_type, handle_model
from akasha.utils.prompts.format import (
    handle_language,
    language_dict,
    handle_metrics,
    handle_params,
    handle_table,
)
import datetime
import logging
from akasha.utils.base import (
    DEFAULT_CHUNK_SIZE,
    DEFAULT_SEARCH_TYPE,
    DEFAULT_EMBED,
    DEFAULT_MODEL,
    DEFAULT_MAX_OUTPUT_TOKENS,
    DEFAULT_MAX_INPUT_TOKENS,
)
from akasha.utils.db.db_structure import dbs
from akasha.utils.db.load_db import process_db, load_db_by_chroma_name


class basic_llm:
    """basic class for akasha, implement _set_model, _change_variables, _check_db, add_log and save_logs function.

    Returns:
        _type_: _description_
    """

    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        max_input_tokens: int = DEFAULT_MAX_INPUT_TOKENS,
        max_output_tokens: int = DEFAULT_MAX_OUTPUT_TOKENS,
        temperature: float = 0.0,
        language: str = "ch",
        record_exp: str = "",
        system_prompt: str = "",
        keep_logs: bool = False,
        verbose: bool = False,
        env_file: str = "",
    ):
        """_summary_

        Args:
            model (str, optional): _description_. Defaults to DEFAULT_MODEL.
            max_input_tokens (int, optional): _description_. Defaults to DEFAULT_MAX_INPUT_TOKENS.
            max_output_tokens (int, optional): _description_. Defaults to DEFAULT_MAX_OUTPUT_TOKENS.
            temperature (float, optional): _description_. Defaults to 0.0.
            language (str, optional): _description_. Defaults to "ch".
            record_exp (str, optional): _description_. Defaults to "".
            system_prompt (str, optional): _description_. Defaults to "".
            keep_logs (bool, optional): _description_. Defaults to False.
            verbose (bool, optional): _description_. Defaults to False.
            env_file (str, optional): _description_. Defaults to "".

        Raises:
            OSError: _description_

        Returns:
            _type_: _description_
        """

        self.verbose = verbose
        self.language = handle_language(language)
        self.record_exp = record_exp
        self.system_prompt = system_prompt
        self.temperature = temperature
        self.keep_logs = keep_logs
        self.max_output_tokens = max_output_tokens
        self.max_input_tokens = max_input_tokens
        self.env_file = env_file
        self.timestamp_list = []
        self.logs: dict[dict] = {}

        ### set model and embeddings ###
        self.model_obj = handle_model(
            model, self.verbose, self.temperature, self.max_output_tokens, self.env_file
        )
        self.model = handle_model_type(model)

    def _set_model(self, **kwargs):
        """change model, temperature if user use **kwargs to change them."""
        ## check if we need to change db, model_obj or embeddings_obj ##

        if (
            ("model" in kwargs)
            or ("temperature" in kwargs)
            or ("max_output_tokens" in kwargs)
            or ("env_file" in kwargs)
        ):
            new_temp = self.temperature
            new_model = self.model
            new_tokens = self.max_output_tokens
            new_env_file = self.env_file
            if "temperature" in kwargs:
                new_temp = kwargs["temperature"]
            if "model" in kwargs:
                new_model = kwargs["model"]
            if "max_output_tokens" in kwargs:
                new_tokens = kwargs["max_output_tokens"]
            if "env_file" in kwargs:
                new_env_file = kwargs["env_file"]
            if (
                (new_model != self.model)
                or (new_temp != self.temperature)
                or (new_tokens != self.max_output_tokens)
                or (new_env_file != self.env_file)
            ):
                self.model_obj = handle_model(
                    new_model, self.verbose, new_temp, new_tokens, new_env_file
                )

    def _change_variables(self, **kwargs):
        """change other arguments if user use **kwargs to change them."""
        ### check input argument is valid or not ###
        for key, value in kwargs.items():
            if (key == "model" or key == "embeddings") and key in self.__dict__:
                self.__dict__[key] = handle_model_type(value)

            elif key == "language":
                self.language = handle_language(value)

            elif key in self.__dict__:  # check if variable exist
                if (
                    getattr(self, key, None) != value
                ):  # check if variable value is different
                    self.__dict__[key] = value
            else:
                if key != "dbs":
                    logging.warning(f"argument {key} not exist")

    def _add_basic_log(self, timestamp: str, fn_type: str) -> bool:
        """add pre-process log to self.logs

        Args:
            timestamp (str): timestamp of this run
            fn_type (str): function type of this run
        """
        if self.keep_logs is False:
            return False
        self.timestamp_list.append(timestamp)

        if timestamp not in self.logs:
            self.logs[timestamp] = {}
        self.logs[timestamp]["fn_type"] = fn_type
        self.logs[timestamp]["model"] = self.model
        self.logs[timestamp]["language"] = language_dict[self.language]
        self.logs[timestamp]["temperature"] = self.temperature
        self.logs[timestamp]["max_input_tokens"] = self.max_input_tokens

        return True

    def _add_result_log(self, timestamp: str, time: float) -> bool:
        """add post-process log to self.logs

        Args:
            timestamp (str): timestamp of this run
            time (float): spent time of this run
        """

        if self.keep_logs is False:
            return False

        self.logs[timestamp]["time"] = time
        try:
            self.logs[timestamp]["docs"] = "\n\n".join(
                [doc.page_content for doc in self.docs]
            )
            self.logs[timestamp]["doc_metadata"] = "\n\n".join(
                [
                    doc.metadata["source"] + "    page: " + str(doc.metadata["page"])
                    for doc in self.docs
                ]
            )
        except Exception:
            try:
                self.logs[timestamp]["doc_metadata"] = "\n\n".join(
                    [
                        doc.metadata["url"] + "    title: " + str(doc.metadata["title"])
                        for doc in self.docs
                    ]
                )
            except Exception:
                try:
                    self.logs[timestamp]["doc_metadata"] = "none"
                    self.logs[timestamp]["docs"] = "\n\n".join(
                        [doc for doc in self.docs]
                    )
                except Exception:
                    self.logs[timestamp]["doc_metadata"] = "none"
                    self.logs[timestamp]["docs"] = "\n\n".join(
                        [doc.page_content for doc in self.docs]
                    )

        self.logs[timestamp]["system_prompt"] = self.system_prompt

        return True

    def _display_docs(self, show_num: int = 5):
        if (not hasattr(self, "docs")) or (self.verbose is False) or (self.docs == []):
            return
        show_less = ""
        if len(self.docs) > show_num:
            show_less = f" (showing {show_num} of " + str(len(self.docs)) + ")"
        splitter = "\n----------------\n"
        found_ref = (
            "----------------\n"
            + splitter.join([doc.page_content for doc in self.docs[:show_num]])
            + splitter
        )

        print(f"Reference{show_less}: \n\n" + found_ref + "\n\n")

        return

    def save_logs(self, file_name: str = "", file_type: str = "json"):
        """save logs into json or txt file

        Args:
            file_name (str, optional): file path and the file name. if not assign, use logs/{current time}. Defaults to "".
            file_type (str, optional): the extension of the file, can be txt or json. Defaults to "json".

        Returns:
            plain_text(str): string of the log
        """

        extension = ""
        ## set extension ##
        if file_name != "":
            tst_file = file_name.split(".")[-1]
            if file_type == "json":
                if tst_file != "json" and tst_file != "JSON":
                    extension = ".json"
            else:
                if tst_file != "txt" and tst_file != "TXT":
                    extension = ".txt"

        ## set filename if not given ##
        from pathlib import Path

        if file_name == "":
            file_name = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

            logs_path = Path("logs")
            if not logs_path.exists():
                logs_path.mkdir()
            file_path = Path("logs/" + file_name + extension)
        else:
            file_path = Path(file_name + extension)

        ## write file ##
        if file_type == "json":
            import json

            with open(file_path, "w", encoding="utf-8") as fp:
                json.dump(self.logs, fp, indent=4, ensure_ascii=False)

        else:
            with open(file_path, "w", encoding="utf-8") as fp:
                for key in self.logs:
                    text = key + ":\n"
                    fp.write(text)
                    for k in self.logs[key]:
                        if type(self.logs[key][k]) is list:
                            text = (
                                k
                                + ": "
                                + "\n".join([str(w) for w in self.logs[key][k]])
                                + "\n\n"
                            )

                        else:
                            text = k + ": " + str(self.logs[key][k]) + "\n\n"

                        fp.write(text)

                    fp.write("\n\n\n\n")

        print("save logs to " + str(file_path))
        return


class atman(basic_llm):
    """basic class for akasha, implement _set_model, _change_variables, _check_db, add_log and save_logs function."""

    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        embeddings: str = DEFAULT_EMBED,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        search_type: Union[str, Callable] = DEFAULT_SEARCH_TYPE,
        max_input_tokens: int = DEFAULT_MAX_INPUT_TOKENS,
        max_output_tokens: int = DEFAULT_MAX_OUTPUT_TOKENS,
        temperature: float = 0.0,
        threshold: float = 0.0,
        language: str = "ch",
        record_exp: str = "",
        system_prompt: str = "",
        keep_logs: bool = False,
        verbose: bool = False,
        use_chroma: bool = False,
        env_file: str = "",
    ):
        """initials of atman class

        Args:
            **chunk_size (int, optional)**: chunk size of texts from documents. Defaults to 1000.\n
            **model (str, optional)**: llm model to use. Defaults to "gpt-3.5-turbo".\n
            embeddings (_type_, optional): embedding model, including two types(openai and huggingface). Defaults to "openai:text-embedding-ada-002".
            **verbose (bool, optional)**: show log texts or not. Defaults to False.\n
            **threshold (float, optional)**: (deprecated) the similarity threshold of searching. Defaults to 0.0.\n
            **language (str, optional)**: the language of documents and prompt, use to make sure docs won't exceed
                max token size of llm input.\n
            **search_type (str, optional)**: search type to find similar documents from db, default 'merge'.
                includes 'merge', 'mmr', 'svm', 'tfidf', also, you can custom your own search_type function, as long as your
                function input is (query_embeds:np.array, docs_embeds:list[np.array], k:int, relevancy_threshold:float, log:dict)
                and output is a list [index of selected documents].\n
            **record_exp (str, optional)**: use aiido to save running params and metrics to the remote mlflow or not if record_exp not empty, and set
                record_exp as experiment name.  default "".\n
            **system_prompt (str, optional)**: the system prompt that you assign special instruction to llm model, so will not be used
                in searching relevant documents. Defaults to "".\n
            **temperature (float, optional)**: temperature of llm model from 0.0 to 1.0 . Defaults to 0.0.\n
            **keep_logs (bool, optional)**: record logs or not. Defaults to False.\n
            **max_output_tokens (int, optional)**: max output tokens of llm model. Defaults to 1024.\n
            **max_input_tokens (int, optional)**: max input tokens of llm model. Defaults to 3000.\n
            **env_file (str, optional)**: the path of the .env file. Defaults to "".\n
        """

        super().__init__(
            model,
            max_input_tokens,
            max_output_tokens,
            temperature,
            language,
            record_exp,
            system_prompt,
            keep_logs,
            verbose,
            env_file,
        )

        self.chunk_size = chunk_size
        self.threshold = threshold
        self.search_type = handle_model_type(search_type, self.verbose)
        self.use_chroma = use_chroma

        self.db = None
        self.ignored_files = []
        ### set model and embeddings ###

        self.embeddings_obj = handle_embeddings(embeddings, self.verbose, self.env_file)
        self.embeddings = handle_model_type(embeddings)

    def _set_model(self, **kwargs):
        """change model, embeddings, search_type, temperature if user use **kwargs to change them."""
        ## check if we need to change db, model_obj or embeddings_obj ##
        super()._set_model(**kwargs)

        if "search_type" in kwargs:
            self.search_type_str = handle_model_type(
                kwargs["search_type"], self.verbose
            )

        if ("embeddings" in kwargs) or ("env_file" in kwargs):
            new_embeddings = self.embeddings
            new_env_file = self.env_file
            if "embeddings" in kwargs:
                new_embeddings = kwargs["embeddings"]
            if "env_file" in kwargs:
                new_env_file = kwargs["env_file"]

            if new_embeddings != self.embeddings or new_env_file != self.env_file:
                self.embeddings_obj = handle_embeddings(
                    new_embeddings, self.verbose, new_env_file
                )

    def _check_doc_path(self, doc_path: Union[List[str], str, dbs]):
        if isinstance(doc_path, dbs):
            return "use dbs object"

        return doc_path

    def _check_db(self) -> bool:
        """check if user input doc_path is exist or not

        Returns:
            _type_: _description_
        """
        if self.db is None:
            info = "document path not exist or don't have any file.\n"
            raise OSError(info)

        return True

    def _get_db(self, data: Union[List[str], str, dbs]):
        if isinstance(data, dbs):
            self.db = data
            self.ignored_files = []
        elif self.use_chroma:
            self.db, self.ignored_files = load_db_by_chroma_name(data)
        else:
            self.db, self.ignored_files = process_db(
                data_source=data,
                embeddings=self.embeddings_obj,
                chunk_size=self.chunk_size,
                verbose=self.verbose,
                env_file=self.env_file,
            )

        return

    def _add_basic_log(self, timestamp: str, fn_type: str) -> bool:
        """add pre-process log to self.logs

        Args:
            timestamp (str): timestamp of this run
            fn_type (str): function type of this run
        """
        if super()._add_basic_log(timestamp, fn_type) is False:
            return False

        self.logs[timestamp]["chunk_size"] = self.chunk_size
        self.logs[timestamp]["embeddings"] = self.embeddings

        return True

    def _add_result_log(self, timestamp: str, time: float):
        """add post-process log to self.logs

        Args:
            timestamp (str): timestamp of this run
            time (float): spent time of this run
        """

        if super()._add_result_log(timestamp, time) is False:
            return False
        self.logs[timestamp]["time"] = time

        return True

    def _upload_logs(
        self, tot_time: float, doc_length: int = 0, doc_tokens: int = 0
    ) -> str:
        if self.record_exp == "":
            return "no record_exp assigned, so no logs uploaded"

        params = handle_params(
            self.model,
            self.language,
            self.search_type,
            self.threshold,
            self.embeddings,
            self.chunk_size,
        )
        metrics = handle_metrics(doc_length, tot_time, doc_tokens)
        table = handle_table(self.prompt, self.docs, self.response)
        from akasha.utils.upload import aiido_upload

        aiido_upload(self.record_exp, params, metrics, table)

        return "logs uploaded"

    def _format_docs(
        self,
    ) -> str:
        splitter = "\n----------------\n"
        found_ref = (
            "----------------\n"
            + splitter.join([doc.page_content for doc in self.docs])
            + splitter
        )

        return found_ref



from langchain_openai import OpenAIEmbeddings, AzureOpenAIEmbeddings
from akasha.utils.prompts.format import language_dict
from langchain_core.embeddings import Embeddings
from typing import Union, Callable, Tuple, List, Dict, Any
from langchain.schema import Document
import jieba
import json
import re
import opencc

jieba.setLogLevel(jieba.logging.INFO)  ## ignore logging jieba model information


cc = opencc.OpenCC("s2twp")


def separate_name(name: str):
    """separate type:name by ':'

    Args:
        **name (str)**: string with format "type:name" \n

    Returns:
        (str, str): res_type , res_name
    """
    sep = name.split(":")

    if len(sep) > 2:
        res_type = sep[0].lower()
        res_name = ":".join(sep[1:])
    elif len(sep) < 2:
        ### if the format type not equal to type:name ###
        res_type = sep[0].lower()
        res_name = ""
    else:
        res_type = sep[0].lower()
        res_name = sep[1]

    return res_type, res_name


def decide_embedding_type(embeddings: Embeddings) -> str:
    """check the embedding type and return the type:name

    Args:
        embeddings (Embeddings): embedding class

    Raises:
        Exception: _description_

    Returns:
        str: type:name
    """
    if isinstance(embeddings, OpenAIEmbeddings) or isinstance(
        embeddings, AzureOpenAIEmbeddings
    ):
        return "openai:" + embeddings.model
    else:
        from akasha.utils.models.gemi import gemini_embed

        if isinstance(embeddings, gemini_embed):
            return "gemini:" + embeddings.model_name

        from akasha.utils.models.custom import custom_embed

        if isinstance(embeddings, custom_embed):
            return embeddings.model_name

        else:
            from langchain_huggingface import HuggingFaceEmbeddings

            if isinstance(embeddings, HuggingFaceEmbeddings):
                return "hf:" + embeddings.model_name

            from langchain_community.embeddings import TensorflowHubEmbeddings

            if isinstance(embeddings, TensorflowHubEmbeddings):
                return "tf:" + embeddings.model_url

            else:
                raise Exception("can not find the embeddings type.")


def get_embedding_type_and_name(
    embeddings: Union[Embeddings, str, Callable],
) -> Tuple[str, str]:
    """get the type and name of the embeddings"""

    if callable(embeddings):
        embeddings_name = embeddings.__name__
    elif isinstance(embeddings, str):
        embeddings_name = embeddings
    else:
        embeddings_name = decide_embedding_type(embeddings)

    embed_type, embed_name = separate_name(embeddings_name)

    return embed_type, embed_name


def sim_to_trad(text: str) -> str:
    """convert simplified chinese to traditional chinese

    Args:
        **text (str)**: simplified chinese\n

    Returns:
        str: traditional chinese
    """
    global cc
    return cc.convert(text)


def get_doc_length(language: str, text: str) -> int:
    """calculate the length of terms in a giving Document

    Args:
        **language (str)**: 'ch' for chinese and 'en' for others, default 'ch'\n
        **doc (Document)**: Document object\n

    Returns:
        doc_length: int Docuemtn length
    """

    if "chinese" in language_dict[language]:
        doc_length = len(list(jieba.cut(text)))
    else:
        doc_length = len(text.split())
    return doc_length


def get_docs_length(language: str, docs: List[Document]) -> int:
    """calculate the total length of terms in giving documents

    Args:
        language (str): 'ch' for chinese and 'en' for others, default 'ch'\n
        docs (list): list of Documents\n

    Returns:
        docs_length: int total Document length
    """
    docs_length = 0
    for doc in docs:
        docs_length += get_doc_length(language, doc.page_content)
    return docs_length


def extract_json(text: str) -> Union[Dict[str, Any], List[Dict[str, Any]], None]:
    """
    Extract JSON data from text generated by LLMs.
    Handles both individual dictionaries and lists of dictionaries.

    Args:
        text (str): The text containing JSON-like data

    Returns:
        Union[Dict[str, Any], List[Dict[str, Any]], None]: Parsed JSON data or None if parsing fails
    """
    # Clean up the text by removing markdown formatting if present
    cleaned_text = re.sub(r"```(?:json)?\n", "", text)
    cleaned_text = re.sub(r"```", "", cleaned_text)

    # Try to extract content within square brackets first (for list of dictionaries)
    list_match = re.search(r"\[(.*?)\]", cleaned_text, re.DOTALL)

    if list_match:
        try:
            # Try to parse as a list of JSON objects
            json_str = list_match.group(0)
            # Fix common issues: replace single quotes with double quotes
            json_str = re.sub(r"'([^']*)':", r'"\1":', json_str)  # Fix keys
            json_str = re.sub(
                r":\s*\'([^\']*)\'", r': "\1"', json_str
            )  # Fix string values
            return json.loads(json_str)
        except json.JSONDecodeError:
            # If list parsing fails, we'll fall through to try single object parsing
            pass

    # Try to extract the most promising JSON object
    # This looks for balanced braces { } with content inside
    matches = []
    stack = []
    start = -1

    for i, char in enumerate(cleaned_text):
        if char == "{":
            if not stack:  # Start of a potential JSON object
                start = i
            stack.append("{")
        elif char == "}":
            if stack and stack[-1] == "{":
                stack.pop()
                if not stack:  # We've found a complete balanced {} section
                    matches.append(cleaned_text[start : i + 1])

    # Process matches from longest to shortest (assuming more complete is better)
    matches.sort(key=len, reverse=True)

    for match in matches:
        try:
            # Fix common issues that LLMs make in JSON generation
            json_str = match
            # Replace single quotes with double quotes for keys and string values
            json_str = re.sub(r"'([^']*)':", r'"\1":', json_str)  # Fix keys
            json_str = re.sub(
                r":\s*\'([^\']*)\'", r': "\1"', json_str
            )  # Fix string values
            # Remove trailing commas before closing brackets
            json_str = re.sub(r",\s*}", "}", json_str)
            json_str = re.sub(r",\s*]", "]", json_str)
            # Fix boolean values
            json_str = re.sub(r":\s*True", r": true", json_str)
            json_str = re.sub(r":\s*False", r": false", json_str)
            # Fix None values
            json_str = re.sub(r":\s*None", r": null", json_str)

            return json.loads(json_str)
        except json.JSONDecodeError:
            continue

    # If all attempts failed, return None
    return None


def extract_multiple_json(text: str) -> List[Dict[str, Any]]:
    """
    Extract multiple JSON objects from text generated by LLMs.

    Args:
        text (str): The text containing multiple JSON objects

    Returns:
        List[Dict[str, Any]]: List of parsed JSON objects
    """
    # Clean up the text
    cleaned_text = re.sub(r"```(?:json)?\n", "", text)
    cleaned_text = re.sub(r"```", "", cleaned_text)

    results = []
    matches = []
    stack = []
    start = -1

    # Find all potential JSON objects with balanced braces
    for i, char in enumerate(cleaned_text):
        if char == "{":
            if not stack:
                start = i
            stack.append("{")
        elif char == "}":
            if stack and stack[-1] == "{":
                stack.pop()
                if not stack:  # Complete balanced object found
                    matches.append(cleaned_text[start : i + 1])

    # Try to parse each potential JSON object
    for match in matches:
        try:
            # Apply the same fixes as in the single object function
            json_str = match
            json_str = re.sub(r"'([^']*)':", r'"\1":', json_str)
            json_str = re.sub(r":\s*\'([^\']*)\'", r': "\1"', json_str)
            json_str = re.sub(r",\s*}", "}", json_str)
            json_str = re.sub(r",\s*]", "]", json_str)
            json_str = re.sub(r":\s*True", r": true", json_str)
            json_str = re.sub(r":\s*False", r": false", json_str)
            json_str = re.sub(r":\s*None", r": null", json_str)

            parsed = json.loads(json_str)
            results.append(parsed)
        except json.JSONDecodeError:
            continue

    return results



import time
from akasha.utils.atman import atman
from akasha.utils.prompts.gen_prompt import (
    default_self_ask_prompt,
    OutputSchema,
    JSON_formatter,
    format_sys_prompt,
)
from akasha.utils.prompts.format import language_dict
from akasha.helper.run_llm import call_model
from akasha.helper.base import extract_json


def self_ask_f(self: atman, start_time: float, timestamp: str) -> str:
    """implement the self ask rag function, first get the follow up questions by user prompt,
    then answer all follow up questions, then use the follow up information to answer the user prompt.

    Args:
        self (atman): _description_
        start_time (float): _description_
        timestamp (str): _description_

    Returns:
        _type_: _description_
    """
    final_prompt = self.prompt

    self_ask_prompt = default_self_ask_prompt() + "Question: " + final_prompt

    formatter = [
        OutputSchema(
            name="need",
            description="the 0 or 1 value that the question need follow up questions to answer or not.",
            type="bool",
        ),
        OutputSchema(
            name="follow_up",
            description="if need follow up questions, list of string of the follow up questions, else return empty list.",
            type="list",
        ),
    ]

    ## format sys prompt ##
    self_ask_sys_prompt = ""
    if "chinese" in language_dict[self.language]:
        self_ask_sys_prompt = "用中文回答 "
    self_ask_sys_prompt += JSON_formatter(formatter)
    prod_sys_prompt = format_sys_prompt(
        self_ask_sys_prompt, self_ask_prompt, self.prompt_format_type, self.model
    )

    ret = call_model(self.model_obj, prod_sys_prompt)
    parse_json = extract_json(ret)

    if parse_json is None or int(parse_json["need"]) == 0:
        self.follow_up = []
        return self(self.db, self.prompt)

    else:
        self.follow_up = parse_json["follow_up"]
    ### start to get response ###

    inter_q = []
    inter_a = []
    tot_prompt_len, tot_prompt_tokens = self.prompt_length, self.prompt_tokens
    tot_doc_len, tot_doc_tokens = 0, 0

    for each_fol_up in self.follow_up:
        each_fol_ans = self(self.db, each_fol_up)
        inter_q.append(each_fol_up)
        inter_a.append(each_fol_ans)
        tot_prompt_len += self.prompt_length
        tot_prompt_tokens += self.prompt_tokens
        tot_doc_len += self.doc_length
        tot_doc_tokens += self.doc_tokens

    inter_info = get_inter_info(inter_q, inter_a)

    text_input = format_sys_prompt(
        self.system_prompt,
        inter_info + "User Question: " + final_prompt,
        self.prompt_format_type,
        self.model,
    )

    self.prompt_length = tot_prompt_len
    self.prompt_tokens = tot_prompt_tokens
    self.doc_length = tot_doc_len
    self.doc_tokens = tot_doc_tokens
    self.prompt = final_prompt

    end_time = time.time()
    if self.stream:
        return self._display_stream(
            text_input,
        )

    self.response = call_model(
        self.model_obj,
        text_input,
    )

    self._add_result_log(timestamp, end_time - start_time)

    self._upload_logs(end_time - start_time, self.doc_length, self.doc_tokens)

    return self.response


def get_inter_info(inter_q: list, inter_a: list) -> str:
    """format the follow up question and answer to a string

    Args:
        inter_q (_type_): _description_
        inter_a (_type_): _description_

    Returns:
        str: _description_
    """
    inter_info = []
    for i in range(len(inter_q)):
        inter_info.append(f"{inter_q[i]}  {inter_a[i]}")
    return (
        "Use follow information to answer User Question\n"
        + "\n----------------------\n".join(inter_info)
    )



from .base import (
    separate_name,
    decide_embedding_type,
    get_embedding_type_and_name,
    get_doc_length,
    get_docs_length,
    sim_to_trad,
    extract_json,
)
from .handle_objects import handle_model_type, handle_embeddings, handle_model
from .token_counter import myTokenizer
from .encoding import detect_encoding, get_mac_address, get_text_md5
from .run_llm import (
    call_model,
    call_batch_model,
    call_stream_model,
    call_image_model,
    call_translator,
    call_JSON_formatter,
)
from .preprocess_prompts import merge_history_and_prompt
from .scores import get_llm_score, get_toxic_score, get_bert_score, get_rouge_score
from .web_engine import load_docs_from_webengine
from .crawler import get_text_from_url
from .self_query_filter import self_query

__all__ = [
    "separate_name",
    "decide_embedding_type",
    "get_embedding_type_and_name",
    "get_doc_length",
    "get_docs_length",
    "sim_to_trad",
    "extract_json",
    "handle_model_type",
    "handle_embeddings",
    "handle_model",
    "myTokenizer",
    "detect_encoding",
    "get_mac_address",
    "get_text_md5",
    "call_model",
    "call_batch_model",
    "call_stream_model",
    "call_image_model",
    "call_translator",
    "call_JSON_formatter",
    "merge_history_and_prompt",
    "load_docs_from_webengine",
    "get_text_from_url",
    "get_llm_score",
    "get_toxic_score",
    "get_bert_score",
    "get_rouge_score",
    "self_query",
]



from typing import List, Union, Set
from akasha.utils.db.db_structure import dbs
import logging
from pathlib import Path


def extract_db_by_file(db: dbs, file_name_list: List[str]) -> dbs:
    """extract db from dbs based on file_name_list

    Args:
        db (dbs): dbs object
        file_name_list (list): list of file names

    Returns:
        dbs: dbs object
    """
    ret_db = dbs()
    file_set = set()
    for file_name in file_name_list:
        file_name = file_name.replace("\\", "/")
        file_name = file_name.lstrip("./")
        file_name = file_name.split("/")[-1]
        file_set.add(file_name)

    for i in range(len(db.ids)):
        mmap = Path(db.metadatas[i]["source"]).name

        if mmap in file_set:
            ret_db.ids.append(db.ids[i])
            ret_db.embeds.append(db.embeds[i])
            ret_db.metadatas.append(db.metadatas[i])
            ret_db.docs.append(db.docs[i])
            ret_db.vis.add(db.ids[i])

    if len(ret_db.ids) == 0:
        logging.warning("No document found.\n\n")
        print("No document found.\n\n")
    return ret_db


def extract_db_by_keyword(db: dbs, keyword_list: List[str]) -> dbs:
    """extract db from dbs based on keyword_list

    Args:
        db (dbs): dbs object
        keyword_list (list): list of keywords

    Returns:
        dbs: dbs object
    """
    ret_db = dbs()
    vis_id = set()

    for keyword in keyword_list:
        for i in range(len(db.ids)):
            if db.ids[i] in vis_id:
                continue
            if keyword in db.docs[i]:
                ret_db.ids.append(db.ids[i])
                ret_db.embeds.append(db.embeds[i])
                ret_db.metadatas.append(db.metadatas[i])
                ret_db.docs.append(db.docs[i])
                ret_db.vis.add(db.ids[i])
                vis_id.add(db.ids[i])

    if len(ret_db.ids) == 0:
        logging.warning("No document found.\n\n")
        print("No document found.\n\n")
    return ret_db


def extract_db_by_ids(db: dbs, id_list: Union[List[str], Set[str]]) -> dbs:
    """extract db from dbs based on ids

    Args:
        db (dbs): dbs object
        id_list (Union[List[str], Set[str]]): list of ids

    Returns:
        dbs: dbs object
    """
    ret_db = dbs()
    if isinstance(id_list, list):
        id_list = set(id_list)

    for idx, id in enumerate(db.ids):
        if id in id_list:
            ret_db.ids.append(db.ids[idx])
            ret_db.embeds.append(db.embeds[idx])
            ret_db.metadatas.append(db.metadatas[idx])
            ret_db.docs.append(db.docs[idx])
            ret_db.vis.add(db.ids[idx])

    if len(ret_db.ids) == 0:
        logging.warning("No document found.\n\n")
        print("No document found.\n\n")
    return ret_db


def pop_db_by_ids(db: dbs, id_list: Union[List[str], Set[str]]):
    """pop undesired data from  dbs based on ids

    Args:
        db (dbs): dbs object
        id_list (Union[List[str], Set[str]]): list of ids

    Returns:
    """
    if isinstance(id_list, list):
        id_list = set(id_list)

    # Filter the lists in place
    db.ids = [id for id in db.ids if id not in id_list]
    db.embeds = [embed for id, embed in zip(db.ids, db.embeds) if id not in id_list]
    db.metadatas = [
        metadata for id, metadata in zip(db.ids, db.metadatas) if id not in id_list
    ]
    db.docs = [doc for id, doc in zip(db.ids, db.docs) if id not in id_list]
    db.vis = {id for id in db.vis if id not in id_list}

    if len(db.ids) == 0:
        logging.warning("No document found.\n\n")
        print("No document found.\n\n")
